unit SXSkinLibrary;

////////////////////////////////////////////////////////////////////////////////
// SXSkinComponents: Skinnable Visual Controls with Alpha Channel             //
//----------------------------------------------------------------------------//
// Version: 1.0.1                                                             //
// Author: Alexey Sadovnikov                                                  //
// Web Site: http://www.saarixx.info                                          //
// E-Mail: mail@saarixx.info                                                  //
//----------------------------------------------------------------------------//
// LICENSE:                                                                   //
// 1. You can freely distribute this file.                                    //
// 2. You can not make any changes to this file.                              //
// 3. The only person who can change this file is Alexey Sadovnikov.          //
// 4. You can use this file in your freeware projects.                        //
// 5. If you want to use this file in your shareware project, you should      //
//    purchase a project license or a personal license of SXSkinComponents.   //
// 6. You can freely use, distribute and modify skins for SXSkinComponents.   //
// 7. You can create skins for SXSkinComponents.                              //
//----------------------------------------------------------------------------//
// Copyright 2006 by Alexey Sadovnikov. All Rights Reserved.                  //
////////////////////////////////////////////////////////////////////////////////

interface

{$I Compilers.inc}

uses Windows, Classes, Forms, SysUtils, IniFiles, GR32, Controls, GR32_Blend,
     SXMathEval, GR32_Polygons, Dialogs, Graphics, Types, SXBitmap32Utils,
     StdCtrls;

type

 {$IFNDEF COMPILER_9_UP}
 TVerticalAlignment=(taAlignTop,
                     taAlignBottom,
                     taVerticalCenter);
 {$ENDIF}

 TSXGlyphPosition=(gpLeftTop,      gpTop,
                   gpRightTop,     gpRight,
                   gpRightBottom,  gpBottom,
                   gpLeftBottom,   gpLeft);

 TSXGlyphChangeAction=(gcaHighlightIn,
                       gcaHighlightOut,
                       gcaDown,
                       gcaUp,
                       gcaCheck,
                       gcaUncheck,
                       gcaEnable,
                       gcaDisable,
                       gcaFocus,
                       gcaUnfocus);

 TSXTransformEffectType=(tetBlend,
                         tetNone,
                         tetClear,
                         tetFade,
                         tetOverDraw,
                         tetSlide);

 TSXTransformEffectDirection=(tedLeft,
                              tedRight,
                              tedTop,
                              tedBottom,
                              tedLeftRight,
                              tedTopBottom,
                              tedLeftTop,
                              tedLeftBottom,
                              tedRightTop,
                              tedRightBottom);

 TSXTransformEffectData=record
  OldType:TSXTransformEffectType;
  OldDirection:TSXTransformEffectDirection;
  OldDirOut:Boolean;
  OldInverted:Boolean;
  OldOnTop:Boolean;
  NewType:TSXTransformEffectType;
  NewDirection:TSXTransformEffectDirection;
  NewDirOut:Boolean;
  NewInverted:Boolean;
  StepsNum:Integer;
  Offset:Integer;
  DrawCaption:Boolean;
 end;

 TSXSkinStyleElement=class
  private
   FControl:TControl;
  public
   Name:String;
   procedure Assign(Element:TSXSkinStyleElement); virtual;
   function GetCopy:TSXSkinStyleElement; virtual;
   procedure SetProperty(const Name,Value,SkinFilePath:String); virtual;
 end;

 TSXSkinStyleElementList=class
  protected
   FItem:TList;
   function Get(Index:Integer):TSXSkinStyleElement;
   procedure Put(Index:Integer;Item:TSXSkinStyleElement);
   function GetCount:Integer;
  public
   procedure Assign(ElementList:TSXSkinStyleElementList);
   function GetIndexByName(const Name:String):Integer;
   procedure Add(SXSkinStyleElement:TSXSkinStyleElement);
   procedure Delete(Index:Integer);
   procedure Clear;
   constructor Create;
   destructor Destroy; override;
   property Item[Index:Integer]:TSXSkinStyleElement read Get write Put; default;
   property Count:Integer read GetCount;
 end;

 TSXSkinStyleImageType=(ssitAutoDetect,
                        ssitJPEG,
                        ssitPNG);

 TSXSkinLibrary=class;

 TSXImageFilterType=(iftNone,
                     iftLighten,
                     iftLightenHorizG,
                     iftLightenVertG,
                     iftDarken,
                     iftDarkenHorizG,
                     iftDarkenVertG,
                     iftAlpha,
                     iftMonochrome,
                     iftColorOverlay,
                     iftColorOverlayHorizG,
                     iftColorOverlayVertG);

 TSXFilterData=record
  SetFilter:Boolean;
  AType:TSXImageFilterType;
  Value:Integer;
  Color:TColor32;
  Color2:TColor32;
 end;

 TSXSkinStyleImageElement=class(TSXSkinStyleElement)
  public
   ImageType:TSXSkinStyleImageType;
   Path:String;
   Transparent:Boolean;
   Bitmap:TBitmap32;
   Filter:TSXFilterData;
   procedure Assign(Element:TSXSkinStyleElement); override;
   function GetCopy:TSXSkinStyleElement; override;
   procedure SetProperty(const Name,Value,SkinFilePath:String); override;
   procedure ValidateBitmap(SkinLibrary:TSXSkinLibrary);
   constructor Create;
   destructor Destroy; override;
 end;

 TSXSkinStyleBoxTileElement=class(TSXSkinStyleElement)
  public
   CenterPath:String;
   LeftPath:String;
   TopPath:String;
   RightPath:String;
   BottomPath:String;
   TopLeftPath:String;
   TopRightPath:String;
   BottomRightPath:String;
   BottomLeftPath:String;
   //
   CenterColor:TColor32;
   LeftColor:TColor32;
   TopColor:TColor32;
   RightColor:TColor32;
   BottomColor:TColor32;
   TopLeftColor:TColor32;
   TopRightColor:TColor32;
   BottomRightColor:TColor32;
   BottomLeftColor:TColor32;
   //
   SetCenterColor:Boolean;
   SetLeftColor:Boolean;
   SetTopColor:Boolean;
   SetRightColor:Boolean;
   SetBottomColor:Boolean;
   SetTopLeftColor:Boolean;
   SetTopRightColor:Boolean;
   SetBottomRightColor:Boolean;
   SetBottomLeftColor:Boolean;
   //
   CenterBitmap:TBitmap32;
   LeftBitmap:TBitmap32;
   TopBitmap:TBitmap32;
   RightBitmap:TBitmap32;
   BottomBitmap:TBitmap32;
   TopLeftBitmap:TBitmap32;
   TopRightBitmap:TBitmap32;
   BottomRightBitmap:TBitmap32;
   BottomLeftBitmap:TBitmap32;
   Transparent:Boolean;
   Filter:TSXFilterData;
   procedure Assign(Element:TSXSkinStyleElement); override;
   function GetCopy:TSXSkinStyleElement; override;   
   procedure DrawToBitmap(B:TBitmap32;X,Y,Width,Height:Integer;DrawRgn:HRGN);
   procedure SetProperty(const Name,Value,SkinFilePath:String); override;
   procedure ValidateBitmaps(SkinLibrary:TSXSkinLibrary);
   constructor Create;
   destructor Destroy; override;
 end;

 TSXSkinStyleFigureType=(ssftRectangle,
                         ssftCircle,
                         ssftRoundRectangle,
                         ssftLine,
                         ssftSolidFill,
                         ssftFocusRectangle);

 TSXFigureControlPreparedData=class
  Control:TControl;
  LastClientRect:TRect;
  Rect:TRect;
  Polygon1:TPolygon32;
  Polygon1Offset:TPoint;
  Polygon2:TPolygon32;
  Polygon2Offset:TPoint;
  destructor Destroy; override;
 end;

 TSXFigureControlPreparedDataList=class
  protected
   FItem:TList;
   function Get(Index:Integer):TSXFigureControlPreparedData;
   procedure Put(Index:Integer;Item:TSXFigureControlPreparedData);
   function GetCount:Integer;
  public
   function GetIndexToInsert(Control:TControl):Integer;
   function GetIndexByControl(Control:TControl):Integer;
   procedure Add(SXFigureControlPreparedData:TSXFigureControlPreparedData);
   procedure Delete(Index:Integer);
   procedure Clear;
   constructor Create;
   destructor Destroy; override;
   property Item[Index:Integer]:TSXFigureControlPreparedData read Get write Put; default;
   property Count:Integer read GetCount;
 end;

 TSXSkinStyleFigureElement=class(TSXSkinStyleElement)
  public
   HasBorder:Boolean;
   HasFill:Boolean;
   FigureType:TSXSkinStyleFigureType;
   BorderColor:TColor32;
   FillColor:TColor32;
   FillColor2:TColor32;
   GradientFill:Boolean;
   VertGradientFill:Boolean;
   Roundness:Integer;
   RectValue:String;
   PolyValue:String;
   CornersStyle:TSXCorners;
   BorderThickness:Single;
   ControlsData:TSXFigureControlPreparedDataList;
   procedure Assign(Element:TSXSkinStyleElement); override;
   function GetCopy:TSXSkinStyleElement; override;   
   procedure DrawToBitmap(B:TBitmap32;X,Y:Integer);
   function OnGetVariable(const VarName:String;var Error:Boolean):Double;
   procedure EvalPrePaintParams(Control:TControl;
              AOnGetVariable:TSXOnGetVariable=nil;Changed:Boolean=False);
   procedure SetProperty(const Name,Value,SkinFilePath:String); override;
   constructor Create;
   destructor Destroy; override;
 end;

 TSXSkinStyle=class
  public
   Name:String;
   procedure AssignStyle(Style:TSXSkinStyle); virtual;
   procedure SetParameter(const Name,Value:String); virtual;
 end;

 TSXSkinGeneralStyle=class(TSXSkinStyle)
  public
   Transparent:Boolean;
   Elements:TSXSkinStyleElementList;
   procedure AssignStyle(Style:TSXSkinStyle); override;
   procedure SetParameter(const Name,Value:String); override;
   constructor Create;
   destructor Destroy; override;
 end;

 TSXFontData=record
  SetFontName:Boolean;
  SetFontSize:Boolean;
  SetFontStyle:Boolean;
  SetFontColor:Boolean;
  SetHasShadow:Boolean;
  SetShadowColor:Boolean;
  SetSmoothLevel:Boolean;
  SetDoPrepaint:Boolean;
  //
  FontName:String;
  FontSize:Integer;
  FontStyle:TFontStyles;
  FontColor:TColor32;
  HasShadow:Boolean;
  ShadowColor:TColor32;
  SmoothLevel:Integer;
  DoPrepaint:Boolean;
 end;

 TSXSkinLabelStateParam=record
  FD:TSXFontData;
 end;
 PSXSkinLabelStateParam=^TSXSkinLabelStateParam;

 TSXSkinLabelStyle=class(TSXSkinStyle)
  public
   NState:TSXSkinLabelStateParam;
   HState:TSXSkinLabelStateParam;
   RState:TSXSkinLabelStateParam;
   procedure AssignStyle(Style:TSXSkinStyle); override;
   procedure SetParameter(const Name,Value:String); override;
 end;

 TSXSkinCheckBoxStateParam=record
  Style:String;
  OverStyle:String;
  GlyphStyle:String;
  GlyphWidth:Integer;
  GlyphHeight:Integer;
  FD:TSXFontData;
  //
  SetCaptionLeftOffset:Boolean;
  SetCaptionTopOffset:Boolean;
  SetCaptionRightOffset:Boolean;
  SetCaptionBottomOffset:Boolean;
  CaptionLeftOffset:Integer;
  CaptionTopOffset:Integer;
  CaptionRightOffset:Integer;
  CaptionBottomOffset:Integer;
  //
  SetTextLeftOffset:Boolean;
  SetTextTopOffset:Boolean;
  SetTextRightOffset:Boolean;
  SetTextBottomOffset:Boolean;
  TextLeftOffset:Integer;
  TextTopOffset:Integer;
  TextRightOffset:Integer;
  TextBottomOffset:Integer;
 end;
 PSXSkinCheckBoxStateParam=^TSXSkinCheckBoxStateParam;

 TSXSkinCheckBoxStyle=class(TSXSkinStyle)
  public
   NUUState:TSXSkinCheckBoxStateParam;
   NFUState:TSXSkinCheckBoxStateParam;
   HUUState:TSXSkinCheckBoxStateParam;
   HFUState:TSXSkinCheckBoxStateParam;
   DFUState:TSXSkinCheckBoxStateParam;
   RUUState:TSXSkinCheckBoxStateParam;
   NUCState:TSXSkinCheckBoxStateParam;
   NFCState:TSXSkinCheckBoxStateParam;
   HUCState:TSXSkinCheckBoxStateParam;
   HFCState:TSXSkinCheckBoxStateParam;
   DFCState:TSXSkinCheckBoxStateParam;
   RUCState:TSXSkinCheckBoxStateParam;
   NUGState:TSXSkinCheckBoxStateParam;
   NFGState:TSXSkinCheckBoxStateParam;
   HUGState:TSXSkinCheckBoxStateParam;
   HFGState:TSXSkinCheckBoxStateParam;
   DFGState:TSXSkinCheckBoxStateParam;
   RUGState:TSXSkinCheckBoxStateParam;
   //
   HInCheckBoxEffect:TSXTransformEffectData;
   HOutCheckBoxEffect:TSXTransformEffectData;
   CheckCheckBoxEffect:TSXTransformEffectData;
   UncheckCheckBoxEffect:TSXTransformEffectData;
   DownCheckBoxEffect:TSXTransformEffectData;
   UpCheckBoxEffect:TSXTransformEffectData;
   EnableCheckBoxEffect:TSXTransformEffectData;
   DisableCheckBoxEffect:TSXTransformEffectData;
   FocusCheckBoxEffect:TSXTransformEffectData;
   UnfocusCheckBoxEffect:TSXTransformEffectData;
   //
   HInGlyphEffect:TSXTransformEffectData;
   HOutGlyphEffect:TSXTransformEffectData;
   CheckGlyphEffect:TSXTransformEffectData;
   UncheckGlyphEffect:TSXTransformEffectData;
   DownGlyphEffect:TSXTransformEffectData;
   UpGlyphEffect:TSXTransformEffectData;
   EnableGlyphEffect:TSXTransformEffectData;
   DisableGlyphEffect:TSXTransformEffectData;
   FocusGlyphEffect:TSXTransformEffectData;
   UnfocusGlyphEffect:TSXTransformEffectData;
   procedure AssignStyle(Style:TSXSkinStyle); override;
   procedure SetParameter(const Name,Value:String); override;
   procedure GetCurrentCBState(var CBState:TSXSkinCheckBoxStateParam;
              State:TCheckBoxState;Down,MouseOver,Focused,Enabled:Boolean);
 end;

 TSXSkinRadioButtonStateParam=record
  Style:String;
  OverStyle:String;
  GlyphStyle:String;
  GlyphWidth:Integer;
  GlyphHeight:Integer;
  FD:TSXFontData;
  //
  SetCaptionLeftOffset:Boolean;
  SetCaptionTopOffset:Boolean;
  SetCaptionRightOffset:Boolean;
  SetCaptionBottomOffset:Boolean;
  CaptionLeftOffset:Integer;
  CaptionTopOffset:Integer;
  CaptionRightOffset:Integer;
  CaptionBottomOffset:Integer;
  //
  SetTextLeftOffset:Boolean;
  SetTextTopOffset:Boolean;
  SetTextRightOffset:Boolean;
  SetTextBottomOffset:Boolean;
  TextLeftOffset:Integer;
  TextTopOffset:Integer;
  TextRightOffset:Integer;
  TextBottomOffset:Integer;
 end;
 PSXSkinRadioButtonStateParam=^TSXSkinRadioButtonStateParam;

 TSXSkinRadioButtonStyle=class(TSXSkinStyle)
  public
   NUUState:TSXSkinRadioButtonStateParam;
   NFUState:TSXSkinRadioButtonStateParam;
   HUUState:TSXSkinRadioButtonStateParam;
   HFUState:TSXSkinRadioButtonStateParam;
   DFUState:TSXSkinRadioButtonStateParam;
   RUUState:TSXSkinRadioButtonStateParam;
   NUCState:TSXSkinRadioButtonStateParam;
   NFCState:TSXSkinRadioButtonStateParam;
   HUCState:TSXSkinRadioButtonStateParam;
   HFCState:TSXSkinRadioButtonStateParam;
   DFCState:TSXSkinRadioButtonStateParam;
   RUCState:TSXSkinRadioButtonStateParam;
   //
   HInRadioButtonEffect:TSXTransformEffectData;
   HOutRadioButtonEffect:TSXTransformEffectData;
   CheckRadioButtonEffect:TSXTransformEffectData;
   UncheckRadioButtonEffect:TSXTransformEffectData;
   DownRadioButtonEffect:TSXTransformEffectData;
   UpRadioButtonEffect:TSXTransformEffectData;
   EnableRadioButtonEffect:TSXTransformEffectData;
   DisableRadioButtonEffect:TSXTransformEffectData;
   FocusRadioButtonEffect:TSXTransformEffectData;
   UnfocusRadioButtonEffect:TSXTransformEffectData;
   //
   HInGlyphEffect:TSXTransformEffectData;
   HOutGlyphEffect:TSXTransformEffectData;
   CheckGlyphEffect:TSXTransformEffectData;
   UncheckGlyphEffect:TSXTransformEffectData;
   DownGlyphEffect:TSXTransformEffectData;
   UpGlyphEffect:TSXTransformEffectData;
   EnableGlyphEffect:TSXTransformEffectData;
   DisableGlyphEffect:TSXTransformEffectData;
   FocusGlyphEffect:TSXTransformEffectData;
   UnfocusGlyphEffect:TSXTransformEffectData;
   procedure AssignStyle(Style:TSXSkinStyle); override;
   procedure SetParameter(const Name,Value:String); override;
   procedure GetCurrentRBState(var RBState:TSXSkinRadioButtonStateParam;
              Checked,Down,MouseOver,Focused,Enabled:Boolean);
 end;

 TSXSkinGroupBoxStateParam=record
  Style:String;
  OverStyle:String;
  SetCaptionPosition:Boolean;
  CaptionPosition:Integer;
  //
  SetCaptionLeftOffset:Boolean;
  SetCaptionTopOffset:Boolean;
  SetCaptionRightOffset:Boolean;
  SetCaptionBottomOffset:Boolean;
  CaptionLeftOffset:Integer;
  CaptionTopOffset:Integer;
  CaptionRightOffset:Integer;
  CaptionBottomOffset:Integer;
  //
  SetTextLeftOffset:Boolean;
  SetTextTopOffset:Boolean;
  SetTextRightOffset:Boolean;
  SetTextBottomOffset:Boolean;
  TextLeftOffset:Integer;
  TextTopOffset:Integer;
  TextRightOffset:Integer;
  TextBottomOffset:Integer;
  //
  TransparentRect:String;
 end;
 PSXSkinGroupBoxStateParam=^TSXSkinGroupBoxStateParam;

 TSXSkinGroupBoxStyle=class(TSXSkinStyle)
  public
   LabelStyle:String;
   NUState:TSXSkinGroupBoxStateParam;
   NFState:TSXSkinGroupBoxStateParam;
   HUState:TSXSkinGroupBoxStateParam;
   HFState:TSXSkinGroupBoxStateParam;
   RUState:TSXSkinGroupBoxStateParam;
   GlyphWidth:Integer;
   GlyphHeight:Integer;
   ClientRect:String;
   procedure AssignStyle(Style:TSXSkinStyle); override;
   procedure SetParameter(const Name,Value:String); override;
   procedure GetCurrentGBState(var GBState:TSXSkinGroupBoxStateParam;
              MouseOver,Focused,Enabled:Boolean);
 end;

 TSXSkinMultiStateStyle=class(TSXSkinStyle)
  public
   NStyle:String;
   HStyle:String;
   DStyle:String;
   RStyle:String;
   procedure AssignStyle(Style:TSXSkinStyle); override;
   procedure SetParameter(const Name,Value:String); override;
 end;

 TSXSkinMultiStateCheckStyle=class(TSXSkinStyle)
  public
   NUStyle:String;
   HUStyle:String;
   DUStyle:String;
   RUStyle:String;
   NCStyle:String;
   HCStyle:String;
   DCStyle:String;
   RCStyle:String;
   procedure AssignStyle(Style:TSXSkinStyle); override;
   procedure SetParameter(const Name,Value:String); override;
 end;

 TSXSkinButtonStateParam=record
  Style:String;
  OverStyle:String;
  FD:TSXFontData;
  SetCaptionLeftOffset:Boolean;
  SetCaptionTopOffset:Boolean;
  SetCaptionRightOffset:Boolean;
  SetCaptionBottomOffset:Boolean;
  CaptionLeftOffset:Integer;
  CaptionTopOffset:Integer;
  CaptionRightOffset:Integer;
  CaptionBottomOffset:Integer;
  DefGlyphFilter:TSXFilterData;
 end;
 PSXSkinButtonStateParam=^TSXSkinButtonStateParam;

 TSXSkinButtonStyle=class(TSXSkinStyle)
  public
   NUUState:TSXSkinButtonStateParam;
   NFUState:TSXSkinButtonStateParam;
   HUUState:TSXSkinButtonStateParam;
   HFUState:TSXSkinButtonStateParam;
   DFUState:TSXSkinButtonStateParam;
   RUUState:TSXSkinButtonStateParam;
   NUCState:TSXSkinButtonStateParam;
   NFCState:TSXSkinButtonStateParam;
   HUCState:TSXSkinButtonStateParam;
   HFCState:TSXSkinButtonStateParam;
   DFCState:TSXSkinButtonStateParam;
   RUCState:TSXSkinButtonStateParam;
   //
   HInButtonEffect:TSXTransformEffectData;
   HOutButtonEffect:TSXTransformEffectData;
   CheckButtonEffect:TSXTransformEffectData;
   UncheckButtonEffect:TSXTransformEffectData;
   DownButtonEffect:TSXTransformEffectData;
   UpButtonEffect:TSXTransformEffectData;
   EnableButtonEffect:TSXTransformEffectData;
   DisableButtonEffect:TSXTransformEffectData;
   FocusButtonEffect:TSXTransformEffectData;
   UnfocusButtonEffect:TSXTransformEffectData;
   //
   HInGlyphEffect:TSXTransformEffectData;
   HOutGlyphEffect:TSXTransformEffectData;
   CheckGlyphEffect:TSXTransformEffectData;
   UncheckGlyphEffect:TSXTransformEffectData;
   DownGlyphEffect:TSXTransformEffectData;
   UpGlyphEffect:TSXTransformEffectData;
   EnableGlyphEffect:TSXTransformEffectData;
   DisableGlyphEffect:TSXTransformEffectData;
   FocusGlyphEffect:TSXTransformEffectData;
   UnfocusGlyphEffect:TSXTransformEffectData;
   //
   HInDDGlyphEffect:TSXTransformEffectData;
   HOutDDGlyphEffect:TSXTransformEffectData;
   CheckDDGlyphEffect:TSXTransformEffectData;
   UncheckDDGlyphEffect:TSXTransformEffectData;
   DownDDGlyphEffect:TSXTransformEffectData;
   UpDDGlyphEffect:TSXTransformEffectData;
   EnableDDGlyphEffect:TSXTransformEffectData;
   DisableDDGlyphEffect:TSXTransformEffectData;
   FocusDDGlyphEffect:TSXTransformEffectData;
   UnfocusDDGlyphEffect:TSXTransformEffectData;
   //
   DropDownGlyphWidth:Integer;
   DropDownGlyphHeight:Integer;
   DropDownGlyphOffset:Integer;
   procedure AssignStyle(Style:TSXSkinStyle); override;
   procedure SetParameter(const Name,Value:String); override;
 end;

 TSXSkinEditStateParam=record
  Style:String;
  FD:TSXFontData;
 end;
 PSXSkinEditStateParam=^TSXSkinEditStateParam;

 TSXSkinEditStyle=class(TSXSkinStyle)
  public
   NUState:TSXSkinEditStateParam;
   NFState:TSXSkinEditStateParam;
   HUState:TSXSkinEditStateParam;
   HFState:TSXSkinEditStateParam;
   RUState:TSXSkinEditStateParam;
   TextLeftOffset:Integer;
   TextTopOffset:Integer;
   TextRightOffset:Integer;
   TextBottomOffset:Integer;
   //
   HInEditEffect:TSXTransformEffectData;
   HOutEditEffect:TSXTransformEffectData;
   EnableEditEffect:TSXTransformEffectData;
   DisableEditEffect:TSXTransformEffectData;
   FocusEditEffect:TSXTransformEffectData;
   UnfocusEditEffect:TSXTransformEffectData;
   procedure AssignStyle(Style:TSXSkinStyle); override;
   procedure SetParameter(const Name,Value:String); override;
 end;

 TSXSkinStyleList=class
  protected
   FItem:TList;
   function Get(Index:Integer):TSXSkinStyle;
   procedure Put(Index:Integer;Item:TSXSkinStyle);
   function GetCount:Integer;
  public
   function GetIndexToInsert(const Name:String):Integer;
   function GetIndexByName(const Name:String):Integer;
   procedure Add(SXSkinStyle:TSXSkinStyle);
   procedure AddUnique(SXSkinStyle:TSXSkinStyle);
   procedure Delete(Index:Integer);
   procedure Clear;
   constructor Create;
   destructor Destroy; override;
   property Item[Index:Integer]:TSXSkinStyle read Get write Put; default;
   property Count:Integer read GetCount;
 end;

 TSXSkinCustomLibrary=class(TComponent)
  private
   FActive:Boolean;
   FSkinDir:String;
   FSkinFile:String;
   FSkinFile2:String;
   FIsLoaded:Boolean;
   Controls:TList;
   Strings:TStringList;
   function GetCanBeUsed:Boolean;
   procedure SetActive(Value:Boolean);
   procedure SetSkinDir(const Value:String);
   procedure SetSkinFile(const Value:String);
   procedure SetSkinFile2(const Value:String);
   function VarValue(const Value:String):String;
   function GetFullSkinFile:String;
   function GetFullSkinFile2:String;
  protected
   procedure Loaded; override; 
  public
   Styles:TSXSkinStyleList;
   procedure AddControl(Control:TControl);
   procedure RemoveControl(Control:TControl);
   procedure InvalidateControls;
   function GetStringsValue(const Name:String):String;
   procedure LoadFromFile(const FilePath:String);
   constructor Create(AOwner:TComponent); override;
   destructor Destroy; override;
   property Active:Boolean read FActive write SetActive;
   property IsLoaded:Boolean read FIsLoaded;
   property CanBeUsed:Boolean read GetCanBeUsed;
   property SkinDir:String read FSkinDir write SetSkinDir;
   property SkinFile:String read FSkinFile write SetSkinFile;
   property SkinFile2:String read FSkinFile2 write SetSkinFile2;
 end;

 TSXSkinLibrary=class(TSXSkinCustomLibrary)
  published
   property Active stored False;
   property SkinDir;
   property SkinFile;
   property SkinFile2;
 end;

procedure ClearFontData(var FD:TSXFontData);
procedure AddFontData(var FD:TSXFontData;FDToAdd:TSXFontData);
function SameFontData(const FD1,FD2:TSXFontData):Boolean;
procedure SetDefaultFontData(var FD:TSXFontData;Font:TFont);
function SameFilterData(const Filter1,Filter2:TSXFilterData):Boolean;
function CreateSkinStyleElementByType(const TypeName:String):TSXSkinStyleElement;
function GetFilterTypeByName(const Name:String):TSXImageFilterType;
function HasTransformEffect(const Effect:TSXTransformEffectData):Boolean;
procedure ApplyFilterToBitmap(var Bitmap:TBitmap32;const Filter:TSXFilterData);
function GetTransformEffectTypeByName(const Name:String):TSXTransformEffectType;
function GetTransformEffectDirectionByName(const Name:String):TSXTransformEffectDirection;
procedure GetTransformEffectByName(const Name:String;var Effect:TSXTransformEffectData);
procedure ApplyTransformEffectToBitmaps(B1,B2:TBitmap32;const Effect:TSXTransformEffectData;
           CurrentStep:Integer;var B:TBitmap32);

procedure GetColorFromString(const S:String;var A,R,G,B:Byte); overload;
function GetColorFromString(const S:String):TColor32; overload;

procedure GetRectFromString(const S:String;var R:TRect;OnGetVar:TSXOnGetVariable=nil);

implementation

uses SXSkinUtils, jpeg, SXPNGUtils, SXSkinControl;

procedure PNGObjectToBitmap32(PNG:TPNGObject;B32:TBitmap32);
var PB:PByte;
    PC:PColor32;
     C:Cardinal;
begin
 B32.SetSize(PNG.Width,PNG.Height);
 if (PNG.Width=0) or (PNG.Height=0) then exit;
 SetStretchBltMode(B32.Canvas.Handle,COLORONCOLOR);
 StretchDiBits(B32.Canvas.Handle,0,0,PNG.Width,PNG.Height,0,0,
               PNG.Width,PNG.Height,PNG.Header.Data,
               pBitmapInfo(@PNG.Header.BitmapInfo)^,DIB_RGB_COLORS,SRCCOPY);
 if PNG.Header.ColorType in [COLOR_GRAYSCALEALPHA,COLOR_RGBALPHA] then
  begin
   PB:=Pointer(PNG.AlphaScanline[0]);
   if PB<>nil then
    begin
     PC:=Pointer(B32.Bits);
     for C:=0 to PNG.Width*PNG.Height-1 do
      begin
       PC^:=SetAlpha(PC^,PB^);
       Inc(PB); Inc(PC);
      end;
    end;
  end else
   begin
    PC:=Pointer(B32.Bits);
    for C:=0 to PNG.Width*PNG.Height-1 do
     begin
      PC^:=SetAlpha(PC^,$FF);
      Inc(PC);
     end;
   end;
end;

procedure GetRectFromString(const S:String;var R:TRect;OnGetVar:TSXOnGetVariable=nil);
var A,B,C,D:Integer;

 procedure SetWhatGot;
 var S2:String;
      E:Boolean;
 begin
  if B>A then
   begin
    S2:=Copy(S,A,B-A);
    if @OnGetVar=nil then
     D:=StrToIntDef(S2,0) else
      begin
       E:=False;
       D:=round(SXEvalMathString(S2,OnGetVar,E));
       if E then D:=0;
      end;
    case C of
     0: R.Left:=D;
     1: R.Top:=D;
     2: R.Right:=D;
     3: R.Bottom:=D;
    end;
   end;
 end;

begin
 R:=Rect(0,0,0,0);
 A:=1; B:=1; C:=0;
 while (B<=length(S)) and (C<4) do
  begin
   if S[B]=',' then
    begin
     SetWhatGot;
     Inc(C); Inc(B);
     A:=B;
    end else Inc(B);
  end;
 SetWhatGot; 
end;

procedure GetPolygonFromString(const S:String;P:TPolygon32;OnGetVar:TSXOnGetVariable=nil);
var A,B,C:Integer;
  LastX,D:Double;

 procedure SetWhatGot;
 var S2:String;
      E:Boolean;
 begin
  if B>A then
   begin
    S2:=Copy(S,A,B-A);
    if @OnGetVar=nil then
     D:=StrToIntDef(S2,0) else
      begin
       E:=False;
       D:=SXEvalMathString(S2,OnGetVar,E);
       if E then D:=0;
      end;
    if C and 1=0 then LastX:=D else
     P.Add(FixedPoint(LastX,D));
    Inc(C);
   end;
 end;

begin
 A:=1; B:=1; C:=0;
 while B<=length(S) do
  begin
   if S[B]=',' then
    begin
     SetWhatGot;
     Inc(B);
     A:=B;
    end else Inc(B);
  end;
 SetWhatGot;
end;

function HexDigit(C:Char):Byte;
begin
 if C in ['A'..'F'] then
  Result:=10+Ord(C)-Ord('A') else
 if C in ['a'..'f'] then
  Result:=10+Ord(C)-Ord('a') else
 if C in ['0'..'9'] then
  Result:=Ord(C)-Ord('0') else
   Result:=0;
end;

procedure GetColorFromString(const S:String;var A,R,G,B:Byte); 
var AA,BB,C,D:Integer;

 procedure SetWhatGot;
 var S2:String;
 begin
  if BB>AA then
   begin
    S2:=Copy(S,AA,BB-AA);
    D:=StrToIntDef(S2,0);
    case C of
     0: A:=D;
     1: R:=D;
     2: G:=D;
     3: B:=D;
    end;
   end;
 end;

 procedure SetAsWinColor(Color:TColor);
 var CL:TColor32;
 begin
  CL:=Color32(Color);
  R:=RedComponent(CL);
  G:=GreenComponent(CL);
  B:=BlueComponent(CL);
 end;

begin
 A:=255; R:=0; G:=0; B:=0;
 if S='' then exit;
 if S='ScrollBar' then
  begin
   SetAsWinColor(clScrollBar); exit;
  end;
 if S='Background' then
  begin
   SetAsWinColor(clBackground); exit;
  end;
 if S='ActiveCaption' then
  begin
   SetAsWinColor(clActiveCaption); exit;
  end;
 if S='InactiveCaption' then
  begin
   SetAsWinColor(clInactiveCaption); exit;
  end;
 if S='Menu' then
  begin
   SetAsWinColor(clMenu); exit;
  end;
 if S='Window' then
  begin
   SetAsWinColor(clWindow); exit;
  end;
 if S='WindowFrame' then
  begin
   SetAsWinColor(clWindowFrame); exit;
  end;
 if S='MenuText' then
  begin
   SetAsWinColor(clMenuText); exit;
  end;
 if S='WindowText' then
  begin
   SetAsWinColor(clWindowText); exit;
  end;
 if S='CaptionText' then
  begin
   SetAsWinColor(clCaptionText); exit;
  end;
 if S='ActiveBorder' then
  begin
   SetAsWinColor(clActiveBorder); exit;
  end;
 if S='InactiveBorder' then
  begin
   SetAsWinColor(clInactiveBorder); exit;
  end;
 if S='AppWorkSpace' then
  begin
   SetAsWinColor(clAppWorkSpace); exit;
  end;
 if S='Highlight' then
  begin
   SetAsWinColor(clHighlight); exit;
  end;
 if S='HighlightText' then
  begin
   SetAsWinColor(clHighlightText); exit;
  end;
 if S='BtnFace' then
  begin
   SetAsWinColor(clBtnFace); exit;
  end;
 if S='BtnShadow' then
  begin
   SetAsWinColor(clBtnShadow); exit;
  end;
 if S='GrayText' then
  begin
   SetAsWinColor(clGrayText); exit;
  end;
 if S='BtnText' then
  begin
   SetAsWinColor(clBtnText); exit;
  end;
 if S='InactiveCaptionText' then
  begin
   SetAsWinColor(clInactiveCaptionText); exit;
  end;
 if S='BtnHighlight' then
  begin
   SetAsWinColor(clBtnHighlight); exit;
  end;
 if S='3DDkShadow' then
  begin
   SetAsWinColor(cl3DDkShadow); exit;
  end;
 if S='3DLight' then
  begin
   SetAsWinColor(cl3DLight); exit;
  end;
 if S='InfoText' then
  begin
   SetAsWinColor(clInfoText); exit;
  end;
 if S='InfoBk' then
  begin
   SetAsWinColor(clInfoBk); exit;
  end;
 {$IFDEF COMPILER_9_UP}
 if S='HotLight' then
  begin
   SetAsWinColor(clHotLight); exit;
  end;
 {$ENDIF}
 if S='GradientActiveCaption' then
  begin
   SetAsWinColor(clGradientActiveCaption); exit;
  end;
 if S='GradientInactiveCaption' then
  begin
   SetAsWinColor(clGradientInactiveCaption); exit;
  end;
 {$IFDEF COMPILER_9_UP}
 if S='MenuHighlight' then
  begin
   SetAsWinColor(clMenuHighlight); exit;
  end;
 if S='MenuBar' then
  begin
   SetAsWinColor(clMenuBar); exit;
  end;
 {$ENDIF}
 if S[1] in ['#','$'] then
  begin
   if (length(S)<>7) and (length(S)<>9) then exit;
   if length(S)=7 then C:=2 else
    begin
     C:=4;
     A:=(HexDigit(S[2]) shl 4) or HexDigit(S[3]);
    end;
   R:=(HexDigit(S[C  ]) shl 4) or HexDigit(S[C+1]);
   G:=(HexDigit(S[C+2]) shl 4) or HexDigit(S[C+3]);
   B:=(HexDigit(S[C+4]) shl 4) or HexDigit(S[C+5]);
   exit;
  end;
 AA:=1; BB:=1; C:=0;
 while (BB<=length(S)) and (C<4) do
  begin
   if S[BB]=',' then
    begin
     SetWhatGot;
     Inc(C); Inc(BB);
     AA:=BB;
    end else Inc(BB);
  end;
 SetWhatGot;
end;

function GetColorFromString(const S:String):TColor32;
var A,R,G,B:Byte;
begin
 GetColorFromString(S,A,R,G,B);
 Result:=Color32(R,G,B,A);
end;

procedure ClearFontData(var FD:TSXFontData);
begin
 Finalize(FD);
 FillChar(FD,sizeof(FD),0);
end;

function SameFontData(const FD1,FD2:TSXFontData):Boolean;
begin
 Result:=(FD1.FontName=FD2.FontName) and (FD1.FontSize=FD2.FontSize) and
         (FD1.FontStyle=FD2.FontStyle) and (FD1.FontColor=FD2.FontColor) and
         (FD1.HasShadow=FD2.HasShadow) and (not FD1.HasShadow or
         (FD1.ShadowColor=FD2.ShadowColor)) and (FD1.DoPrepaint=FD2.DoPrepaint) and
         (FD1.SmoothLevel=FD2.SmoothLevel);
end;

procedure AddFontData(var FD:TSXFontData;FDToAdd:TSXFontData);
begin
 if not FD.SetFontName then
  begin
   FD.SetFontName:=FDToAdd.SetFontName;
   FD.FontName:=FDToAdd.FontName;
  end;
 if not FD.SetFontSize then
  begin
   FD.SetFontSize:=FDToAdd.SetFontSize;
   FD.FontSize:=FDToAdd.FontSize;
  end;
 if not FD.SetFontStyle then
  begin
   FD.SetFontStyle:=FDToAdd.SetFontStyle;
   FD.FontStyle:=FDToAdd.FontStyle;
  end;
 if not FD.SetFontColor then
  begin
   FD.SetFontColor:=FDToAdd.SetFontColor;
   FD.FontColor:=FDToAdd.FontColor;
  end;
 if not FD.SetHasShadow then
  begin
   FD.SetHasShadow:=FDToAdd.SetHasShadow;
   FD.HasShadow:=FDToAdd.HasShadow;
  end;
 if not FD.SetShadowColor then
  begin
   FD.SetShadowColor:=FDToAdd.SetShadowColor;
   FD.ShadowColor:=FDToAdd.ShadowColor;
  end;
 if not FD.SetSmoothLevel then
  begin
   FD.SetSmoothLevel:=FDToAdd.SetSmoothLevel;
   FD.SmoothLevel:=FDToAdd.SmoothLevel;
  end;
 if not FD.SetDoPrepaint then
  begin
   FD.SetDoPrepaint:=FDToAdd.SetDoPrepaint;
   FD.DoPrepaint:=FDToAdd.DoPrepaint;
  end;
end;

procedure SetDefaultFontData(var FD:TSXFontData;Font:TFont);
begin
 if not FD.SetFontName then
  begin
   FD.FontName:=Font.Name;
   FD.SetFontName:=True;
  end;
 if not FD.SetFontSize then
  begin
   FD.FontSize:=Font.Size;
   FD.SetFontSize:=True;
  end;
 if not FD.SetFontStyle then
  begin
   FD.FontStyle:=Font.Style;
   FD.SetFontStyle:=True;
  end;
 if not FD.SetFontColor then
  begin
   FD.FontColor:=Color32(Font.Color);
   FD.SetFontColor:=True;
  end; 
end;

function TryToSetFontDataParameter(const Name,Value:String;var FD:TSXFontData):Boolean;
var C:Integer;
begin
 Result:=False;
 if Name='FontName' then
  begin
   FD.FontName:=WithoutAllSpaces(Value);
   FD.SetFontName:=True;
   Result:=True;
  end else
 if Name='FontSize' then
  begin
   FD.FontSize:=StrToIntDef(Value,0);
   FD.SetFontSize:=True;
   Result:=True;
  end else
 if Name='FontColor' then
  begin
   FD.FontColor:=GetColorFromString(Value);
   FD.SetFontColor:=True;
   Result:=True;
  end else
 if Name='HasTextShadow' then
  begin
   FD.HasShadow:=Value='1';
   FD.SetHasShadow:=True;
   Result:=True;
  end else
 if Name='ShadowColor' then
  begin
   FD.ShadowColor:=GetColorFromString(Value);
   FD.SetShadowColor:=True;
   Result:=True;
  end else
 if Name='SmoothLevel' then
  begin
   FD.SmoothLevel:=StrToIntDef(Value,0);
   FD.SetSmoothLevel:=True;
   Result:=True;
  end else
 if Name='DoPrepaint' then
  begin
   FD.DoPrepaint:=Value='1';
   FD.SetDoPrepaint:=True;
   Result:=True;
  end else
 if Name='FontStyle' then
  begin
   for C:=1 to length(Value) do
    case Value[C] of
     'B','b': FD.FontStyle:=FD.FontStyle+[fsBold];
     'I','i': FD.FontStyle:=FD.FontStyle+[fsItalic];
     'U','u': FD.FontStyle:=FD.FontStyle+[fsUnderline];
     'S','s': FD.FontStyle:=FD.FontStyle+[fsStrikeOut];
    end;
   FD.SetFontStyle:=True;
   Result:=True; 
  end;
end;

function SameFilterData(const Filter1,Filter2:TSXFilterData):Boolean;
begin
 Result:=((Filter1.AType=iftNone) and (Filter2.AType=iftNone)) or
         ((Filter1.AType=Filter2.AType) and (Filter1.Color=Filter2.Color) and
          (Filter1.Color2=Filter2.Color2) and (Filter1.Value=Filter2.Value));
end;

function TryToSetFilterDataParameter(const Name,Value:String;var Filter:TSXFilterData;
           const Prefix:String=''):Boolean;
var AName:String;
begin
 Result:=False;
 if (Prefix<>'') and (Copy(Name,1,length(Prefix))<>Prefix) then exit;
 if Prefix='' then AName:=Name else
  AName:=Copy(Name,length(Prefix)+1,MaxInt);
 if AName='Filter' then
  begin
   Filter.SetFilter:=True;
   Filter.AType:=GetFilterTypeByName(Value);
   Result:=True;
  end else
 if (AName='FilterColor') or (AName='FilterColor1') then
  begin
   Filter.Color:=GetColorFromString(Value);
   if Filter.AType in [iftLighten,iftLightenHorizG,iftLightenVertG,iftDarken,
    iftDarkenHorizG,iftDarkenVertG] then Filter.Color:=Filter.Color and $00FFFFFF;
   Result:=True;
  end else
 if AName='FilterColor2' then
  begin
   Filter.Color2:=GetColorFromString(Value);
   if Filter.AType in [iftLighten,iftLightenHorizG,iftLightenVertG,iftDarken,
    iftDarkenHorizG,iftDarkenVertG] then Filter.Color2:=Filter.Color2 and $00FFFFFF;
   Result:=True;
  end else
 if AName='FilterValue' then
  begin
   Filter.Value:=StrToIntDef(Value,0);
   Result:=True;
  end;
end;

function TryToSetTransformEffectDataParameter(const Name,Value:String;
           var Effect:TSXTransformEffectData;const Prefix:String=''):Boolean;
var AName:String;
begin
 Result:=False;
 if (Prefix<>'') and (Copy(Name,1,length(Prefix))<>Prefix) then exit;
 if Prefix='' then AName:=Name else
  AName:=Copy(Name,length(Prefix)+1,MaxInt);
 if AName='Transform' then
  begin
   GetTransformEffectByName(Value,Effect);
   Result:=True;
  end else
 if AName='TransformO' then
  begin
   Effect.OldType:=GetTransformEffectTypeByName(Value);
   Result:=True;
  end else
 if AName='TransformDirO' then
  begin
   Effect.OldDirection:=GetTransformEffectDirectionByName(Value);
   Result:=True;
  end else
 if AName='TransformOutO' then
  begin
   Effect.OldDirOut:=Value='1';
   Result:=True;
  end else
 if AName='TransformInvertO' then
  begin
   Effect.OldInverted:=Value='1';
   Result:=True;
  end else
 if (AName='Transform') or (AName='TransformN') then
  begin
   Effect.NewType:=GetTransformEffectTypeByName(Value);
   Result:=True;
  end else
 if AName='TransformDirN' then
  begin
   Effect.NewDirection:=GetTransformEffectDirectionByName(Value);
   Result:=True;
  end else
 if AName='TransformOutN' then
  begin
   Effect.NewDirOut:=Value='1';
   Result:=True;
  end else
 if AName='TransformInvertN' then
  begin
   Effect.NewInverted:=Value='1';
   Result:=True;
  end else
 if AName='TransformOffset' then
  begin
   Effect.Offset:=StrToIntDef(Value,0);
   Result:=True;
  end else
 if AName='TransformOldOnTop' then
  begin
   Effect.OldOnTop:=Value='1';
   Result:=True;
  end else
 if AName='TransformDrawCaption' then
  begin
   Effect.DrawCaption:=Value='1';
   Result:=True;
  end else
 if AName='StepsNum' then
  begin
   Effect.StepsNum:=StrToIntDef(Value,0);
   Result:=True;
  end;
end;

{ TSXSkinStyleElement }

procedure TSXSkinStyleElement.Assign(Element:TSXSkinStyleElement);
begin
 Name:=Element.Name;
end;

function TSXSkinStyleElement.GetCopy:TSXSkinStyleElement;
begin
 Result:=TSXSkinStyleElement.Create;
 Result.Assign(Self); 
end;

procedure TSXSkinStyleElement.SetProperty(const Name,Value,SkinFilePath:String);
begin
end;

{ TSXSkinStyleImageElement }

procedure TSXSkinStyleImageElement.Assign(Element:TSXSkinStyleElement);
begin
 inherited;
 if Element is TSXSkinStyleImageElement then
  begin
   ImageType:=TSXSkinStyleImageElement(Element).ImageType;
   Path:=TSXSkinStyleImageElement(Element).Path;
   Transparent:=TSXSkinStyleImageElement(Element).Transparent;
   Filter:=TSXSkinStyleImageElement(Element).Filter;
   Bitmap.Free;
  end;
end;

function TSXSkinStyleImageElement.GetCopy:TSXSkinStyleElement;
begin
 Result:=TSXSkinStyleImageElement.Create;
 Result.Assign(Self);
end;

procedure TSXSkinStyleImageElement.ValidateBitmap(SkinLibrary:TSXSkinLibrary);
var JPG:TJpegImage;
    PNG:TPNGObject;
begin
 if Bitmap=nil then
  begin
   if ImageType=ssitAutoDetect then
    begin
     if SameText(ExtractFileExt(Path),'.png') then
      ImageType:=ssitPNG else ImageType:=ssitJPEG;
    end;
   case ImageType of
    ssitJPEG: begin
               JPG:=TJpegImage.Create;
               try
                try
                 JPG.LoadFromFile(Path);
                except end;
                Bitmap:=TBitmap32.Create;
                Bitmap.Assign(JPG);
               finally
                JPG.Free;
               end;
              end;
    ssitPNG:  begin
               PNG:=TPNGObject.Create;
               try
                try
                 PNG.LoadFromFile(Path);
                except end;
                Bitmap:=TBitmap32.Create;
                PNGObjectToBitmap32(PNG,Bitmap);
                if (PNG.Width>0) and (PNG.Header.ColorType in [COLOR_GRAYSCALEALPHA,COLOR_RGBALPHA]) then
                 begin
                  Bitmap.DrawMode:=dmBlend;
                  Bitmap.CombineMode:=cmMerge;
                 end; 
               finally
                PNG.Free;
               end;
              end;
   end;
   ApplyFilterToBitmap(Bitmap,Filter);
  end;
end;

procedure TSXSkinStyleImageElement.SetProperty(const Name,Value,SkinFilePath:String);
begin
 if Name='Transparent' then Transparent:=Value='1' else
  if Name='Path' then Path:=GetFullPath(Value,SkinFilePath) else
   TryToSetFilterDataParameter(Name,Value,Filter);
end;

constructor TSXSkinStyleImageElement.Create;
begin
 inherited Create;
 Transparent:=True;
end;

destructor TSXSkinStyleImageElement.Destroy;
begin
 Bitmap.Free;
 inherited Destroy;
end;

{ TSXSkinStyleBoxTileElement }

procedure TSXSkinStyleBoxTileElement.Assign(Element:TSXSkinStyleElement);
begin
 inherited;
 if Element is TSXSkinStyleBoxTileElement then
  begin
   CenterPath:=TSXSkinStyleBoxTileElement(Element).CenterPath;
   LeftPath:=TSXSkinStyleBoxTileElement(Element).LeftPath;
   TopPath:=TSXSkinStyleBoxTileElement(Element).TopPath;
   RightPath:=TSXSkinStyleBoxTileElement(Element).RightPath;
   BottomPath:=TSXSkinStyleBoxTileElement(Element).BottomPath;
   TopLeftPath:=TSXSkinStyleBoxTileElement(Element).TopLeftPath;
   TopRightPath:=TSXSkinStyleBoxTileElement(Element).TopRightPath;
   BottomRightPath:=TSXSkinStyleBoxTileElement(Element).BottomRightPath;
   BottomLeftPath:=TSXSkinStyleBoxTileElement(Element).BottomLeftPath;
   //
   CenterColor:=TSXSkinStyleBoxTileElement(Element).CenterColor;
   LeftColor:=TSXSkinStyleBoxTileElement(Element).LeftColor;
   TopColor:=TSXSkinStyleBoxTileElement(Element).TopColor;
   RightColor:=TSXSkinStyleBoxTileElement(Element).RightColor;
   BottomColor:=TSXSkinStyleBoxTileElement(Element).BottomColor;
   TopLeftColor:=TSXSkinStyleBoxTileElement(Element).TopLeftColor;
   TopRightColor:=TSXSkinStyleBoxTileElement(Element).TopRightColor;
   BottomRightColor:=TSXSkinStyleBoxTileElement(Element).BottomRightColor;
   BottomLeftColor:=TSXSkinStyleBoxTileElement(Element).BottomLeftColor;
   //
   SetCenterColor:=TSXSkinStyleBoxTileElement(Element).SetCenterColor;
   SetLeftColor:=TSXSkinStyleBoxTileElement(Element).SetLeftColor;
   SetTopColor:=TSXSkinStyleBoxTileElement(Element).SetTopColor;
   SetRightColor:=TSXSkinStyleBoxTileElement(Element).SetRightColor;
   SetBottomColor:=TSXSkinStyleBoxTileElement(Element).SetBottomColor;
   SetTopLeftColor:=TSXSkinStyleBoxTileElement(Element).SetTopLeftColor;
   SetTopRightColor:=TSXSkinStyleBoxTileElement(Element).SetTopRightColor;
   SetBottomRightColor:=TSXSkinStyleBoxTileElement(Element).SetBottomRightColor;
   SetBottomLeftColor:=TSXSkinStyleBoxTileElement(Element).SetBottomLeftColor;
   //
   CenterBitmap.Free;
   LeftBitmap.Free;
   TopBitmap.Free;
   RightBitmap.Free;
   BottomBitmap.Free;
   TopLeftBitmap.Free;
   TopRightBitmap.Free;
   BottomRightBitmap.Free;
   BottomLeftBitmap.Free;
   //
   Transparent:=TSXSkinStyleBoxTileElement(Element).Transparent;
   Filter:=TSXSkinStyleBoxTileElement(Element).Filter;
  end;
end;

function TSXSkinStyleBoxTileElement.GetCopy:TSXSkinStyleElement;
begin
 Result:=TSXSkinStyleBoxTileElement.Create;
 Result.Assign(Self);
end;

procedure TSXSkinStyleBoxTileElement.DrawToBitmap(B:TBitmap32;X,Y,Width,Height:Integer;DrawRgn:HRGN);
var LeftW,CenterW,RightW:Integer;
    TopH,CenterH,BottomH:Integer;
                       R:TRect;
              BB,TBitmap:TBitmap32;
                   TX,TY:Integer;

 procedure SetSizeParam(Bitmap:TBitmap32;UseColor:Boolean;var AWidth,AHeight:Integer);
 begin
  if UseColor then
   begin
    if AWidth<1 then AWidth:=1;
    if AHeight<1 then AHeight:=1;
   end else
  if Bitmap<>nil then
   begin
    if Bitmap.Width>AWidth then
     AWidth:=Bitmap.Width;
    if Bitmap.Height>AHeight then
     AHeight:=Bitmap.Height;
   end;
 end;

 procedure DrawBitmapToRect(Bitmap:TBitmap32;UseColor:Boolean;Color:TColor32;R:TRect);
 var  TX2,TY2:Integer;
  DRect,SRect:TRect;
          C,D:Integer;
 begin
  if RectInRegion(DrawRgn,R) then
   begin
    if UseColor then
     begin
      OffsetRect(R,X,Y);
      B.FillRectTS(R,Color);
     end else
    if (Bitmap<>nil) and (Bitmap.Width>0) and (Bitmap.Height>0) then
     begin
      if (Bitmap.Width>=R.Right-R.Left) and (Bitmap.Height>=R.Bottom-R.Top) then
       B.Draw(X+R.Left,Y+R.Top,Bitmap) else
      if (Bitmap.Height=1) and (Bitmap.Width=R.Right-R.Left) then
       begin
        OffsetRect(R,X,Y);
        Bitmap.DrawTo(B,R);
       end else
      if (Bitmap.Width=1) and (Bitmap.Height=R.Bottom-R.Top) then
       begin
        OffsetRect(R,X,Y);
        Bitmap.DrawTo(B,R);
       end else
      if (Bitmap.Width=1) and (Bitmap.Height=1) then
       begin
        OffsetRect(R,X,Y);
        B.FillRectTS(R,Bitmap.Pixel[0,0]);
       end else
        begin
         TX2:=(R.Right-R.Left) div Bitmap.Width;
         if (R.Right-R.Left) mod Bitmap.Width<>0 then Inc(TX2);
         TY2:=(R.Bottom-R.Top) div Bitmap.Height;
         if (R.Bottom-R.Top) mod Bitmap.Height<>0 then Inc(TY2);
         for C:=0 to TY2 do
          for D:=0 to TX2 do
           begin
            SRect:=Bitmap.BoundsRect;
            DRect:=SRect;
            OffsetRect(DRect,D*Bitmap.Width+R.Left,C*Bitmap.Height+R.Top);
            if R.Left>DRect.Left then
             begin
              Inc(SRect.Left,R.Left-DRect.Left);
              DRect.Left:=R.Left;
             end;
            if R.Top>DRect.Top then
             begin
              Inc(SRect.Top,R.Top-DRect.Top);
              DRect.Top:=R.Top;
             end;
            if R.Right<DRect.Right then
             begin
              Dec(SRect.Right,DRect.Right-R.Right);
              DRect.Right:=R.Right;
             end;
            if R.Bottom<DRect.Bottom then
             begin
              Dec(SRect.Bottom,DRect.Bottom-R.Bottom);
              DRect.Bottom:=R.Bottom;
             end;
            OffsetRect(DRect,X,Y);
            Bitmap.DrawTo(B,DRect,SRect);
           end;
        end;
     end;
   end;
 end;

begin
 if Filter.AType<>iftNone then
  begin
   GetRgnBox(DrawRgn,R);
   IntersectRect(R,R,Rect(0,0,Width,Height));
   BB:=TBitmap32.Create;
   BB.SetSize(R.Right-R.Left,R.Bottom-R.Top);
   BB.Clear(0);
   BB.DrawMode:=dmBlend;
   BB.CombineMode:=cmMerge;
   TBitmap:=B;
   B:=BB;
   TX:=X; TY:=Y;
   X:=-R.Left; Y:=-R.Top;
  end else
   begin
    BB:=nil; TBitmap:=nil;
    TX:=0; TY:=0;
   end;
 LeftW:=0; RightW:=0;
 TopH:=0;  BottomH:=0;
 SetSizeParam(TopLeftBitmap,SetTopLeftColor,LeftW,TopH);
 SetSizeParam(TopRightBitmap,SetTopRightColor,RightW,TopH);
 SetSizeParam(BottomLeftBitmap,SetBottomLeftColor,LeftW,BottomH);
 SetSizeParam(BottomRightBitmap,SetBottomRightColor,RightW,BottomH);
 CenterW:=Width-LeftW-RightW;
 CenterH:=Height-TopH-BottomH;
 if CenterW<0 then
  begin
   Inc(RightW,CenterW);
   CenterW:=0;
   if RightW<0 then
    begin
     Inc(LeftW,RightW);
     RightW:=0;
    end;
  end;
 if CenterH<0 then
  begin
   Inc(BottomH,CenterH);
   CenterH:=0;
   if BottomH<0 then
    begin
     Inc(TopH,BottomH);
     BottomH:=0;
    end;
  end;
 if TopH>0 then
  begin
   if LeftW>0 then
    DrawBitmapToRect(TopLeftBitmap,SetTopLeftColor,TopLeftColor,Rect(0,0,LeftW,TopH));
   if CenterW>0 then
    DrawBitmapToRect(TopBitmap,SetTopColor,TopColor,Rect(LeftW,0,Width-RightW,TopH));
   if RightW>0 then
    DrawBitmapToRect(TopRightBitmap,SetTopRightColor,TopRightColor,Rect(Width-RightW,0,Width,TopH));
  end;
 if CenterH>0 then
  begin
   if LeftW>0 then
    DrawBitmapToRect(LeftBitmap,SetLeftColor,LeftColor,Rect(0,TopH,LeftW,Height-BottomH));
   if CenterW>0 then
    DrawBitmapToRect(CenterBitmap,SetCenterColor,CenterColor,Rect(LeftW,TopH,Width-RightW,Height-BottomH));
   if RightW>0 then
    DrawBitmapToRect(RightBitmap,SetRightColor,RightColor,Rect(Width-RightW,TopH,Width,Height-BottomH));
  end;
 if BottomH>0 then
  begin
   if LeftW>0 then
    DrawBitmapToRect(BottomLeftBitmap,SetBottomLeftColor,BottomLeftColor,Rect(0,Height-BottomH,LeftW,Height));
   if CenterW>0 then
    DrawBitmapToRect(BottomBitmap,SetBottomColor,BottomColor,Rect(LeftW,Height-BottomH,Width-RightW,Height));
   if RightW>0 then
    DrawBitmapToRect(BottomRightBitmap,SetBottomRightColor,BottomRightColor,Rect(Width-RightW,Height-BottomH,Width,Height));
  end;
 if Filter.AType<>iftNone then
  begin
   ApplyFilterToBitmap(BB,Filter);
   TBitmap.Draw(TX-X,TY-Y,BB);
   BB.Free;
  end; 
end;

procedure TSXSkinStyleBoxTileElement.ValidateBitmaps(SkinLibrary:TSXSkinLibrary);

 procedure LoadSXBitmap(const Path:String;var Bitmap:TBitmap32);
 var T:TSXSkinStyleImageType;
   JPG:TJpegImage;
   PNG:TPNGObject;
 begin
  if (Bitmap=nil) and (Path<>'') then
   begin
    if SameText(ExtractFileExt(Path),'.png') then
     T:=ssitPNG else T:=ssitJPEG;
    case T of
     ssitJPEG: begin
                JPG:=TJpegImage.Create;
                try
                 try
                  JPG.LoadFromFile(Path);
                 except end;
                 Bitmap:=TBitmap32.Create;
                 Bitmap.Assign(JPG);
                finally
                 JPG.Free;
                end;
               end;
     ssitPNG:  begin
                PNG:=TPNGObject.Create;
                try
                 try
                  PNG.LoadFromFile(Path);
                 except end;
                 Bitmap:=TBitmap32.Create;
                 PNGObjectToBitmap32(PNG,Bitmap);
                 if (PNG.Width>0) and (PNG.Header.ColorType in [COLOR_GRAYSCALEALPHA,COLOR_RGBALPHA]) then
                  begin
                   Bitmap.DrawMode:=dmBlend;
                   Bitmap.CombineMode:=cmMerge;
                  end;
                finally
                 PNG.Free;
                end;
               end;
    end;
   end;
 end;

begin
 if not SetCenterColor then
  LoadSXBitmap(CenterPath,CenterBitmap);
 if not SetLeftColor then
  LoadSXBitmap(LeftPath,LeftBitmap);
 if not SetTopColor then
  LoadSXBitmap(TopPath,TopBitmap);
 if not SetRightColor then
  LoadSXBitmap(RightPath,RightBitmap);
 if not SetBottomColor then
  LoadSXBitmap(BottomPath,BottomBitmap);
 if not SetTopLeftColor then
 LoadSXBitmap(TopLeftPath,TopLeftBitmap);
 if not SetTopRightColor then
  LoadSXBitmap(TopRightPath,TopRightBitmap);
 if not SetBottomRightColor then
  LoadSXBitmap(BottomRightPath,BottomRightBitmap);
 if not SetBottomLeftColor then
  LoadSXBitmap(BottomLeftPath,BottomLeftBitmap);
end;

procedure TSXSkinStyleBoxTileElement.SetProperty(const Name,Value,SkinFilePath:String);
var T:Integer;

 procedure ProcessPathOrColor(var Path:String;var Color:TColor32;var UseColor:Boolean);
 begin
  if (Value<>'') and (Value[1]='#') then
   begin
    Color:=GetColorFromString(Value);
    UseColor:=True;
   end else
    begin
     Path:=GetFullPath(Value,SkinFilePath);
     UseColor:=False;
    end;
 end;

 procedure ProcessMaskPath(const PathPart:String;var Path:String;var UseColor:Boolean);
 var S:String;
 begin
  S:=Copy(Value,1,T-1)+PathPart+Copy(Value,T+1,MaxInt);
  S:=GetFullPath(S,SkinFilePath);
  if FileExists(S) then
   begin
    Path:=S;
    UseColor:=False;
   end; 
 end;

begin
 if Name='Transparent' then Transparent:=Value='1' else
  if Name='Paths' then
   begin
    T:=Pos('*',Value);
    if T>0 then
     begin
      ProcessMaskPath('center',CenterPath,SetCenterColor);
      ProcessMaskPath('left',LeftPath,SetLeftColor);
      ProcessMaskPath('top',TopPath,SetTopColor);
      ProcessMaskPath('right',RightPath,SetRightColor);
      ProcessMaskPath('bottom',BottomPath,SetBottomColor);
      ProcessMaskPath('topleft',TopLeftPath,SetTopLeftColor);
      ProcessMaskPath('topright',TopRightPath,SetTopRightColor);
      ProcessMaskPath('bottomright',BottomRightPath,SetBottomRightColor);
      ProcessMaskPath('bottomleft',BottomLeftPath,SetBottomLeftColor);
     end;
   end else
  if Name='Center' then ProcessPathOrColor(CenterPath,CenterColor,SetCenterColor) else
  if Name='Left' then ProcessPathOrColor(LeftPath,LeftColor,SetLeftColor) else
  if Name='Top' then ProcessPathOrColor(TopPath,TopColor,SetTopColor) else
  if Name='Right' then ProcessPathOrColor(RightPath,RightColor,SetRightColor) else
  if Name='Bottom' then ProcessPathOrColor(BottomPath,BottomColor,SetBottomColor) else
  if Name='TopLeft' then ProcessPathOrColor(TopLeftPath,TopLeftColor,SetTopLeftColor) else
  if Name='TopRight' then ProcessPathOrColor(TopRightPath,TopRightColor,SetTopRightColor) else
  if Name='BottomRight' then ProcessPathOrColor(BottomRightPath,BottomRightColor,SetBottomRightColor) else
  if Name='BottomLeft' then ProcessPathOrColor(BottomLeftPath,BottomLeftColor,SetBottomLeftColor) else
   TryToSetFilterDataParameter(Name,Value,Filter);
end;

constructor TSXSkinStyleBoxTileElement.Create;
begin
 inherited Create;
 Transparent:=True;
end;

destructor TSXSkinStyleBoxTileElement.Destroy;
begin
 CenterBitmap.Free;
 LeftBitmap.Free;
 TopBitmap.Free;
 RightBitmap.Free;
 BottomBitmap.Free;
 TopLeftBitmap.Free;
 TopRightBitmap.Free;
 BottomRightBitmap.Free;
 BottomLeftBitmap.Free;
 inherited Destroy;
end;

{ TSXFigureControlPreparedData }

destructor TSXFigureControlPreparedData.Destroy;
begin
 Polygon1.Free;
 Polygon2.Free;
 inherited Destroy;
end;

{ TSXFigureControlPreparedDataList }

function TSXFigureControlPreparedDataList.Get(Index:Integer):TSXFigureControlPreparedData;
begin
 Result:=TSXFigureControlPreparedData(FItem[Index]);
end;

procedure TSXFigureControlPreparedDataList.Put(Index:Integer;Item:TSXFigureControlPreparedData);
begin
 FItem[Index]:=Item;
end;

function TSXFigureControlPreparedDataList.GetCount:Integer;
begin
 Result:=FItem.Count;
end;

function TSXFigureControlPreparedDataList.GetIndexByControl(Control:TControl):Integer;
var A,L,H:Integer;
        B:Boolean;
begin
 B:=False;
 L:=0;
 H:=Count-1;
 while L<=H do
  begin
   A:=(L+H) shr 1;
   if Cardinal(Item[A].Control)<Cardinal(Control) then L:=A+1 else
    begin
     H:=A-1;
     if Cardinal(Item[A].Control)=Cardinal(Control) then
      begin
       L:=A;
       B:=True;
      end;
    end;
  end;
 if B then Result:=L else Result:=-1;
end;

function TSXFigureControlPreparedDataList.GetIndexToInsert(Control:TControl):Integer;
var A,L,H:Integer;
begin
 L:=0;
 H:=Count-1;
 while L<=H do
  begin
   A:=(L+H) shr 1;
   if Cardinal(Item[A].Control)<Cardinal(Control) then L:=A+1 else
    begin
     H:=A-1;
     if Cardinal(Item[A].Control)=Cardinal(Control) then L:=A;
    end;
  end;
 Result:=L;
end;

procedure TSXFigureControlPreparedDataList.Add(SXFigureControlPreparedData:TSXFigureControlPreparedData);
var A:Integer;
begin
 A:=GetIndexToInsert(SXFigureControlPreparedData.Control);
 FItem.Insert(A,SXFigureControlPreparedData);
end;

procedure TSXFigureControlPreparedDataList.Delete(Index:Integer);
begin
 Item[Index].Free;
 FItem.Delete(Index);
end;

procedure TSXFigureControlPreparedDataList.Clear;
var A:Integer;
begin
 for A:=0 to Count-1 do
  Item[A].Free;
 FItem.Clear;
end;

constructor TSXFigureControlPreparedDataList.Create;
begin
 inherited Create;
 FItem:=TList.Create;
end;

destructor TSXFigureControlPreparedDataList.Destroy;
begin
 Clear;
 FItem.Free;
 inherited Destroy;
end;

{ TSXSkinStyleFigureElement }

procedure TSXSkinStyleFigureElement.Assign(Element:TSXSkinStyleElement);
begin
 inherited;
 if Element is TSXSkinStyleFigureElement then
  begin
   HasBorder:=TSXSkinStyleFigureElement(Element).HasBorder;
   HasFill:=TSXSkinStyleFigureElement(Element).HasFill;
   FigureType:=TSXSkinStyleFigureElement(Element).FigureType;
   BorderColor:=TSXSkinStyleFigureElement(Element).BorderColor;
   FillColor:=TSXSkinStyleFigureElement(Element).FillColor;
   FillColor2:=TSXSkinStyleFigureElement(Element).FillColor2;
   GradientFill:=TSXSkinStyleFigureElement(Element).GradientFill;
   VertGradientFill:=TSXSkinStyleFigureElement(Element).VertGradientFill;
   Roundness:=TSXSkinStyleFigureElement(Element).Roundness;
   RectValue:=TSXSkinStyleFigureElement(Element).RectValue;
   PolyValue:=TSXSkinStyleFigureElement(Element).PolyValue;
   CornersStyle:=TSXSkinStyleFigureElement(Element).CornersStyle;
   BorderThickness:=TSXSkinStyleFigureElement(Element).BorderThickness;
   ControlsData.Clear;
  end; 
end;

function TSXSkinStyleFigureElement.GetCopy:TSXSkinStyleElement;
begin
 Result:=TSXSkinStyleFigureElement.Create;
 Result.Assign(Self);
end;

procedure TSXSkinStyleFigureElement.DrawToBitmap(B:TBitmap32;X,Y:Integer);
var T:TSXFigureControlPreparedData;
    A:Integer;
begin
 A:=ControlsData.GetIndexByControl(FControl);
 if A<0 then exit;
 T:=ControlsData[A];
 case FigureType of
  ssftRectangle:      begin
                       if (T.Polygon1<>nil) and ((T.Polygon1Offset.X<>X) or (T.Polygon1Offset.Y<>Y)) then
                        begin
                         T.Polygon1.Offset(Fixed(X-T.Polygon1Offset.X),Fixed(Y-T.Polygon1Offset.Y));
                         T.Polygon1Offset.X:=X;
                         T.Polygon1Offset.Y:=Y;
                        end;
                       if (T.Polygon2<>nil) and ((T.Polygon2Offset.X<>X) or (T.Polygon2Offset.Y<>Y)) then
                        begin
                         T.Polygon2.Offset(Fixed(X-T.Polygon2Offset.X),Fixed(Y-T.Polygon2Offset.Y));
                         T.Polygon2Offset.X:=X;
                         T.Polygon2Offset.Y:=Y;
                        end;
                       if HasFill then
                        begin
                         if GradientFill then
                          begin
                           if VertGradientFill then
                            RectVFadeT(B,FillColor,FillColor2,X+T.Rect.Left,Y+T.Rect.Top,T.Rect.Right-T.Rect.Left,T.Rect.Bottom-T.Rect.Top+1) else
                             RectHFadeT(B,FillColor,FillColor2,X+T.Rect.Left,Y+T.Rect.Top,T.Rect.Right-T.Rect.Left+1,T.Rect.Bottom-T.Rect.Top);
                          end else
                           B.FillRectTS(X+T.Rect.Left,Y+T.Rect.Top,X+T.Rect.Right,Y+T.Rect.Bottom+1,FillColor);
                        end;
                       if HasBorder then
                        begin
                         if BorderThickness=0 then
                          T.Polygon1.DrawEdge(B,BorderColor) else
                           begin
                            T.Polygon2.DrawFill(B,BorderColor);
                           end;
                        end;
                      end;
  ssftCircle:         begin
                       if (T.Polygon1<>nil) and ((T.Polygon1Offset.X<>X) or (T.Polygon1Offset.Y<>Y)) then
                        begin
                         T.Polygon1.Offset(Fixed(X-T.Polygon1Offset.X),Fixed(Y-T.Polygon1Offset.Y));
                         T.Polygon1Offset.X:=X;
                         T.Polygon1Offset.Y:=Y;
                        end;
                       if (T.Polygon2<>nil) and ((T.Polygon2Offset.X<>X) or (T.Polygon2Offset.Y<>Y)) then
                        begin
                         T.Polygon2.Offset(Fixed(X-T.Polygon2Offset.X),Fixed(Y-T.Polygon2Offset.Y));
                         T.Polygon2Offset.X:=X;
                         T.Polygon2Offset.Y:=Y;
                        end;
                       if HasFill then
                        begin
                         if GradientFill then
                          begin
                           if VertGradientFill then
                            CircleVFade(B,FillColor,FillColor2,X+T.Rect.Left,Y+T.Rect.Top,T.Rect.Right-T.Rect.Left-1,T.Rect.Bottom-T.Rect.Top) else
                             CircleHFade(B,FillColor,FillColor2,X+T.Rect.Left,Y+T.Rect.Top,T.Rect.Right-T.Rect.Left-1,T.Rect.Bottom-T.Rect.Top-1);
                          end else CircleFill(B,FillColor,X+T.Rect.Left,Y+T.Rect.Top,T.Rect.Right-T.Rect.Left-1,T.Rect.Bottom-T.Rect.Top);
                        end;
                       if HasBorder then
                        begin
                         if BorderThickness=0 then
                          T.Polygon1.DrawEdge(B,BorderColor) else
                           begin
                            T.Polygon2.DrawFill(B,BorderColor);
                           end;
                        end;
                      end;
  ssftRoundRectangle: begin
                       if (T.Polygon1<>nil) and ((T.Polygon1Offset.X<>X) or (T.Polygon1Offset.Y<>Y)) then
                        begin
                         T.Polygon1.Offset(Fixed(X-T.Polygon1Offset.X),Fixed(Y-T.Polygon1Offset.Y));
                         T.Polygon1Offset.X:=X;
                         T.Polygon1Offset.Y:=Y;
                        end;
                       if (T.Polygon2<>nil) and ((T.Polygon2Offset.X<>X) or (T.Polygon2Offset.Y<>Y)) then
                        begin
                         T.Polygon2.Offset(Fixed(X-T.Polygon2Offset.X),Fixed(Y-T.Polygon2Offset.Y));
                         T.Polygon2Offset.X:=X;
                         T.Polygon2Offset.Y:=Y;
                        end;
                       if HasFill then
                        begin
                         if GradientFill then
                          begin
                           if VertGradientFill then
                            RoundRectVFadeT(B,FillColor,FillColor2,X+T.Rect.Left,Y+T.Rect.Top,T.Rect.Right-T.Rect.Left,T.Rect.Bottom-T.Rect.Top+1,Roundness,CornersStyle) else
                             RoundRectHFadeT(B,FillColor,FillColor2,X+T.Rect.Left,Y+T.Rect.Top,T.Rect.Right-T.Rect.Left+1,T.Rect.Bottom-T.Rect.Top,Roundness,CornersStyle);
                          end else
                           RoundRectFill(B,FillColor,X+T.Rect.Left,Y+T.Rect.Top,T.Rect.Right-T.Rect.Left,T.Rect.Bottom-T.Rect.Top+1,Roundness);
                        end;
                       if HasBorder then
                        begin
                         if BorderThickness=0 then
                          begin
                            T.Polygon1.DrawEdge(B,BorderColor);
                          end else
                           begin
                            T.Polygon2.DrawFill(B,BorderColor);
                           end;
                        end;
                      end;
  ssftLine:           begin
                       if (T.Polygon1<>nil) and ((T.Polygon1Offset.X<>X) or (T.Polygon1Offset.Y<>Y)) then
                        begin
                         T.Polygon1.Offset(Fixed(X-T.Polygon1Offset.X),Fixed(Y-T.Polygon1Offset.Y));
                         T.Polygon1Offset.X:=X;
                         T.Polygon1Offset.Y:=Y;
                        end;
                       if (T.Polygon2<>nil) and ((T.Polygon2Offset.X<>X) or (T.Polygon2Offset.Y<>Y)) then
                        begin
                         T.Polygon2.Offset(Fixed(X-T.Polygon2Offset.X),Fixed(Y-T.Polygon2Offset.Y));
                         T.Polygon2Offset.X:=X;
                         T.Polygon2Offset.Y:=Y;
                        end;
                       if BorderThickness=0 then
                        T.Polygon1.DrawEdge(B,BorderColor) else
                         begin
                          T.Polygon2.DrawFill(B,BorderColor);
                         end;
                      end;
  ssftFocusRectangle: begin
                       OffsetRect(T.Rect,X,Y);
                       DrawFocusRect(B.Handle,T.Rect);
                       OffsetRect(T.Rect,-X,-Y);
                      end;
 end;
end;

function TSXSkinStyleFigureElement.OnGetVariable(const VarName:String;var Error:Boolean):Double;
begin
 Result:=0;
 if FControl<>nil then
  begin
   if VarName='W' then
    begin
     Result:=FControl.Width;
     exit;
    end;
   if VarName='H' then
    begin
     Result:=FControl.Height;
     exit;
    end;
  end;
 Error:=True;
end;

procedure TSXSkinStyleFigureElement.EvalPrePaintParams(Control:TControl;
           AOnGetVariable:TSXOnGetVariable=nil;Changed:Boolean=False);
var A:Integer;
    T:TSXFigureControlPreparedData;
   TP:TPolygon32;
begin
 if @AOnGetVariable=nil then
  AOnGetVariable:=OnGetVariable;
 FControl:=Control;
 A:=ControlsData.GetIndexByControl(Control);
 if A<0 then
  begin
   T:=TSXFigureControlPreparedData.Create;
   T.Control:=Control;
  end else T:=ControlsData[A];
 if not EqualRect(T.LastClientRect,Rect(0,0,Control.Width,Control.Height)) or Changed then
  begin
   T.LastClientRect:=Rect(0,0,Control.Width,Control.Height);
   if RectValue<>'' then
    GetRectFromString(RectValue,T.Rect,AOnGetVariable);

   T.Polygon1.Free;
   T.Polygon2.Free;
   T.Polygon1:=nil;
   T.Polygon2:=nil;
   if HasFill or HasBorder then
    begin
     T.Polygon1:=TPolygon32.Create;
     case FigureType of
      ssftRectangle:      SetRectangle(T.Polygon1,T.Rect.Left,T.Rect.Top,
                            T.Rect.Right-T.Rect.Left,T.Rect.Bottom-T.Rect.Top);
      ssftCircle:         SetCircle(T.Polygon1,(T.Rect.Left+T.Rect.Right-1)/2,
                            (T.Rect.Top+T.Rect.Bottom-1)/2,
                            (T.Rect.Right-T.Rect.Left-1)/2,
                            (T.Rect.Bottom-T.Rect.Top-1)/2);
      ssftRoundRectangle: SetRoundRectangle(T.Polygon1,T.Rect.Left,T.Rect.Top,
                            T.Rect.Right-T.Rect.Left,T.Rect.Bottom-T.Rect.Top,
                            Roundness,CornersStyle);
      ssftLine:           begin
                           GetPolygonFromString(PolyValue,T.Polygon1,AOnGetVariable);
                           T.Polygon1.Closed:=False;
                          end;
     end;
     T.Polygon1.Antialiased:=True;
     T.Polygon1.AntialiasMode:=am4times;
     T.Polygon1Offset.X:=0;
     T.Polygon1Offset.Y:=0;
    end;
   if HasBorder and (BorderThickness<>0) then
    begin
     TP:=T.Polygon1.Outline;
     T.Polygon2:=TP.Grow(Fixed(BorderThickness/2),0.5);
     T.Polygon2.FillMode:=pfWinding;
     TP.Free;
     T.Polygon2Offset.X:=0;
     T.Polygon2Offset.Y:=0;
    end;
  end;
 if A<0 then ControlsData.Add(T); 
end;

procedure TSXSkinStyleFigureElement.SetProperty(const Name,Value,SkinFilePath:String);
begin
 if Name='HasBorder' then HasBorder:=Value='1' else
  if Name='HasFill' then HasFill:=Value='1' else
  if Name='Rect' then RectValue:=WithoutAllSpaces(Value) else
  if Name='Poly' then PolyValue:=WithoutAllSpaces(Value) else
  if Name='Fill' then
   begin
    VertGradientFill:=Value='GRADIENT_V';
    GradientFill:=VertGradientFill or (Value='GRADIENT_H');
   end else
  if Name='BorderColor' then
   BorderColor:=GetColorFromString(Value) else
  if (Name='FillColor') or (Name='FillColor1') then
   FillColor:=GetColorFromString(Value) else
  if Name='FillColor2' then
   FillColor2:=GetColorFromString(Value) else
  if Name='Roundness' then Roundness:=StrToIntDef(Value,0) else
  if Name='CornersStyle' then Byte(CornersStyle):=StrToIntDef(Value,0) else
  if Name='BorderThickness' then BorderThickness:=SXStrToFloatDef(Value,0);
end;

constructor TSXSkinStyleFigureElement.Create;
begin
 inherited Create;
 ControlsData:=TSXFigureControlPreparedDataList.Create;
 CornersStyle:=[crLeftTop,crRightTop,crRightBottom,crLeftBottom];
end;

destructor TSXSkinStyleFigureElement.Destroy;
begin
 ControlsData.Free;
 inherited Destroy;
end;

{ TSXSkinStyle }

procedure TSXSkinStyle.AssignStyle(Style:TSXSkinStyle);
begin
end;

procedure TSXSkinStyle.SetParameter(const Name,Value:String);
begin
end;

{ TSXSkinGeneralStyle }

procedure TSXSkinGeneralStyle.AssignStyle(Style:TSXSkinStyle);
begin
 inherited;
 if Style is TSXSkinGeneralStyle then
  begin
   Elements.Assign(TSXSkinGeneralStyle(Style).Elements);
   Transparent:=TSXSkinGeneralStyle(Style).Transparent;
  end;
end;

procedure TSXSkinGeneralStyle.SetParameter(const Name,Value:String);
begin
end;

constructor TSXSkinGeneralStyle.Create;
begin
 inherited Create;
 Elements:=TSXSkinStyleElementList.Create;
 Transparent:=True;
end;

destructor TSXSkinGeneralStyle.Destroy;
begin
 Elements.Free;
 inherited Destroy;
end;

{ TSXSkinStyleList }

function TSXSkinStyleList.Get(Index:Integer):TSXSkinStyle;
begin
 Result:=TSXSkinStyle(FItem[Index]);
end;

procedure TSXSkinStyleList.Put(Index:Integer;Item:TSXSkinStyle);
begin
 FItem[Index]:=Item;
end;

function TSXSkinStyleList.GetCount:Integer;
begin
 Result:=FItem.Count;
end;

function TSXSkinStyleList.GetIndexByName(const Name:String):Integer;
label l1,l2,l3;
var A,L,H:Integer;
        B:Boolean;
begin
 B:=False;
 L:=0;
 H:=Count-1;
 while L<=H do
  begin
   A:=(L+H) shr 1;
   SXLStrCmp(Item[A].Name,Name);
   asm
    jz l2;
    jnb l1;
   end;
   L:=A+1;
   goto l3;
l2: L:=A; B:=True;
l1: H:=A-1; l3:
  end;
 if B then Result:=L else Result:=-1;
end;

function TSXSkinStyleList.GetIndexToInsert(const Name:String):Integer;
label l1,l2,l3;
var A,L,H:Integer;
begin
 L:=0;
 H:=Count-1;
 while L<=H do
  begin
   A:=(L+H) shr 1;
   SXLStrCmp(Item[A].Name,Name);
   asm
    jz l2;
    jnb l1;
   end;
   L:=A+1;
   goto l3;
l2: L:=A;
l1: H:=A-1; l3:
  end;
 Result:=L;
end;

procedure TSXSkinStyleList.Add(SXSkinStyle:TSXSkinStyle);
var A:Integer;
begin
 A:=GetIndexToInsert(SXSkinStyle.Name);
 FItem.Insert(A,SXSkinStyle);
end;

procedure TSXSkinStyleList.AddUnique(SXSkinStyle:TSXSkinStyle);
var A:Integer;
begin
 A:=GetIndexByName(SXSkinStyle.Name);
 if A<0 then
  begin
   A:=GetIndexToInsert(SXSkinStyle.Name);
   FItem.Insert(A,SXSkinStyle);
  end else
   begin
    Item[A].Free;
    FItem[A]:=SXSkinStyle;
   end;
end;

procedure TSXSkinStyleList.Delete(Index:Integer);
begin
 Item[Index].Free;
 FItem.Delete(Index);
end;

procedure TSXSkinStyleList.Clear;
var A:Integer;
begin
 for A:=0 to Count-1 do
  Item[A].Free;
 FItem.Clear;
end;

constructor TSXSkinStyleList.Create;
begin
 inherited Create;
 FItem:=TList.Create;
end;

destructor TSXSkinStyleList.Destroy;
begin
 Clear;
 FItem.Free;
 inherited Destroy;
end;

{ TSXSkinStyleElementList }

function TSXSkinStyleElementList.Get(Index:Integer):TSXSkinStyleElement;
begin
 Result:=TSXSkinStyleElement(FItem[Index]);
end;

procedure TSXSkinStyleElementList.Put(Index:Integer;Item:TSXSkinStyleElement);
begin
 FItem[Index]:=Item;
end;

function TSXSkinStyleElementList.GetCount:Integer;
begin
 Result:=FItem.Count;
end;

function TSXSkinStyleElementList.GetIndexByName(const Name:String):Integer;
begin
 Result:=Count-1;
 while (Result>0) and (Item[Result].Name<>Name) do Dec(Result);
end;

procedure TSXSkinStyleElementList.Assign(ElementList:TSXSkinStyleElementList);
var A:Integer;
begin
 Clear;
 for A:=0 to ElementList.Count-1 do
  Add(ElementList[A].GetCopy);
end;

procedure TSXSkinStyleElementList.Add(SXSkinStyleElement:TSXSkinStyleElement);
begin
 FItem.Add(SXSkinStyleElement);
end;

procedure TSXSkinStyleElementList.Delete(Index:Integer);
begin
 Item[Index].Free;
 FItem.Delete(Index);
end;

procedure TSXSkinStyleElementList.Clear;
var A:Integer;
begin
 for A:=0 to Count-1 do
  Item[A].Free;
 FItem.Clear;
end;

constructor TSXSkinStyleElementList.Create;
begin
 inherited Create;
 FItem:=TList.Create;
end;

destructor TSXSkinStyleElementList.Destroy;
begin
 Clear;
 FItem.Free;
 inherited Destroy;
end;

{ TSXSkinCheckBoxStyle }

procedure TSXSkinCheckBoxStyle.AssignStyle(Style:TSXSkinStyle);
begin
 inherited;
 if Style is TSXSkinCheckBoxStyle then
  begin
   NUUState:=TSXSkinCheckBoxStyle(Style).NUUState;
   NFUState:=TSXSkinCheckBoxStyle(Style).NFUState;
   HUUState:=TSXSkinCheckBoxStyle(Style).HUUState;
   HFUState:=TSXSkinCheckBoxStyle(Style).HFUState;
   DFUState:=TSXSkinCheckBoxStyle(Style).DFUState;
   RUUState:=TSXSkinCheckBoxStyle(Style).RUUState;
   NUCState:=TSXSkinCheckBoxStyle(Style).NUCState;
   NFCState:=TSXSkinCheckBoxStyle(Style).NFCState;
   HUCState:=TSXSkinCheckBoxStyle(Style).HUCState;
   HFCState:=TSXSkinCheckBoxStyle(Style).HFCState;
   DFCState:=TSXSkinCheckBoxStyle(Style).DFCState;
   RUCState:=TSXSkinCheckBoxStyle(Style).RUCState;
   NUGState:=TSXSkinCheckBoxStyle(Style).NUGState;
   NFGState:=TSXSkinCheckBoxStyle(Style).NFGState;
   HUGState:=TSXSkinCheckBoxStyle(Style).HUGState;
   HFGState:=TSXSkinCheckBoxStyle(Style).HFGState;
   DFGState:=TSXSkinCheckBoxStyle(Style).DFGState;
   RUGState:=TSXSkinCheckBoxStyle(Style).RUGState;
   //
   HInCheckBoxEffect:=TSXSkinCheckBoxStyle(Style).HInCheckBoxEffect;
   HOutCheckBoxEffect:=TSXSkinCheckBoxStyle(Style).HOutCheckBoxEffect;
   CheckCheckBoxEffect:=TSXSkinCheckBoxStyle(Style).CheckCheckBoxEffect;
   UncheckCheckBoxEffect:=TSXSkinCheckBoxStyle(Style).UncheckCheckBoxEffect;
   DownCheckBoxEffect:=TSXSkinCheckBoxStyle(Style).DownCheckBoxEffect;
   UpCheckBoxEffect:=TSXSkinCheckBoxStyle(Style).UpCheckBoxEffect;
   EnableCheckBoxEffect:=TSXSkinCheckBoxStyle(Style).EnableCheckBoxEffect;
   DisableCheckBoxEffect:=TSXSkinCheckBoxStyle(Style).DisableCheckBoxEffect;
   FocusCheckBoxEffect:=TSXSkinCheckBoxStyle(Style).FocusCheckBoxEffect;
   UnfocusCheckBoxEffect:=TSXSkinCheckBoxStyle(Style).UnfocusCheckBoxEffect;
   //
   HInGlyphEffect:=TSXSkinButtonStyle(Style).HInGlyphEffect;
   HOutGlyphEffect:=TSXSkinButtonStyle(Style).HOutGlyphEffect;
   CheckGlyphEffect:=TSXSkinButtonStyle(Style).CheckGlyphEffect;
   UncheckGlyphEffect:=TSXSkinButtonStyle(Style).UncheckGlyphEffect;
   DownGlyphEffect:=TSXSkinButtonStyle(Style).DownGlyphEffect;
   UpGlyphEffect:=TSXSkinButtonStyle(Style).UpGlyphEffect;
   EnableGlyphEffect:=TSXSkinButtonStyle(Style).EnableGlyphEffect;
   DisableGlyphEffect:=TSXSkinButtonStyle(Style).DisableGlyphEffect;
   FocusGlyphEffect:=TSXSkinButtonStyle(Style).FocusGlyphEffect;
   UnfocusGlyphEffect:=TSXSkinButtonStyle(Style).UnfocusGlyphEffect;
  end;
end;

procedure TSXSkinCheckBoxStyle.SetParameter(const Name,Value:String);
var CBParam:PSXSkinCheckBoxStateParam;
          S:String;
begin
 CBParam:=nil;
 S:=Copy(Name,1,4);
 if S='NUU_' then CBParam:=@NUUState else
  if S='NFU_' then CBParam:=@NFUState else
  if S='HUU_' then CBParam:=@HUUState else
  if S='HFU_' then CBParam:=@HFUState else
  if S='DFU_' then CBParam:=@DFUState else
  if S='RUU_' then CBParam:=@RUUState else
  if S='NUC_' then CBParam:=@NUCState else
  if S='NFC_' then CBParam:=@NFCState else
  if S='HUC_' then CBParam:=@HUCState else
  if S='HFC_' then CBParam:=@HFCState else
  if S='DFC_' then CBParam:=@DFCState else
  if S='RUC_' then CBParam:=@RUCState else
  if S='NUG_' then CBParam:=@NUGState else
  if S='NFG_' then CBParam:=@NFGState else
  if S='HUG_' then CBParam:=@HUGState else
  if S='HFG_' then CBParam:=@HFGState else
  if S='DFG_' then CBParam:=@DFGState else
  if S='RUG_' then CBParam:=@RUGState;
 if CBParam=nil then
  begin
   if not TryToSetTransformEffectDataParameter(Name,Value,HInCheckBoxEffect,'HighlightInCheckBox') then
    if not TryToSetTransformEffectDataParameter(Name,Value,HOutCheckBoxEffect,'HighlightOutCheckBox') then
    if not TryToSetTransformEffectDataParameter(Name,Value,CheckCheckBoxEffect,'CheckCheckBox') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UncheckCheckBoxEffect,'UncheckCheckBox') then
    if not TryToSetTransformEffectDataParameter(Name,Value,DownCheckBoxEffect,'DownCheckBox') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UpCheckBoxEffect,'UpCheckBox') then
    if not TryToSetTransformEffectDataParameter(Name,Value,EnableCheckBoxEffect,'EnableCheckBox') then
    if not TryToSetTransformEffectDataParameter(Name,Value,DisableCheckBoxEffect,'DisableCheckBox') then
    if not TryToSetTransformEffectDataParameter(Name,Value,FocusCheckBoxEffect,'FocusCheckBox') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UnfocusCheckBoxEffect,'UnfocusCheckBox') then
    //
    if not TryToSetTransformEffectDataParameter(Name,Value,HInGlyphEffect,'HighlightInGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,HOutGlyphEffect,'HighlightOutGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,CheckGlyphEffect,'CheckGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UncheckGlyphEffect,'UncheckGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,DownGlyphEffect,'DownGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UpGlyphEffect,'UpGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,EnableGlyphEffect,'EnableGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,DisableGlyphEffect,'DisableGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,FocusGlyphEffect,'FocusGlyph') then
     TryToSetTransformEffectDataParameter(Name,Value,UnfocusGlyphEffect,'UnfocusGlyph');
  end else
   begin
    S:=Copy(Name,5,MaxInt);
    if S='Style' then CBParam.Style:=Value else
     if S='OverStyle' then CBParam.OverStyle:=Value else
     if S='Glyph' then CBParam.GlyphStyle:=Value else
     if S='GlyphWidth' then CBParam.GlyphWidth:=StrToIntDef(Value,0) else
     if S='GlyphHeight' then CBParam.GlyphHeight:=StrToIntDef(Value,0) else
     if S='CaptionLeftOffset' then
      begin
       CBParam.SetCaptionLeftOffset:=True;
       CBParam.CaptionLeftOffset:=StrToIntDef(Value,0);
      end else
     if S='CaptionTopOffset' then
      begin
       CBParam.SetCaptionTopOffset:=True;
       CBParam.CaptionTopOffset:=StrToIntDef(Value,0);
      end else
     if S='CaptionRightOffset' then
      begin
       CBParam.SetCaptionRightOffset:=True;
       CBParam.CaptionRightOffset:=StrToIntDef(Value,0);
      end else
     if S='CaptionBottomOffset' then
      begin
       CBParam.SetCaptionBottomOffset:=True;
       CBParam.CaptionBottomOffset:=StrToIntDef(Value,0);
      end else
     if S='TextLeftOffset' then
      begin
       CBParam.SetTextLeftOffset:=True;
       CBParam.TextLeftOffset:=StrToIntDef(Value,0);
      end else
     if S='TextTopOffset' then
      begin
       CBParam.SetTextTopOffset:=True;
       CBParam.TextTopOffset:=StrToIntDef(Value,0);
      end else
     if S='TextRightOffset' then
      begin
       CBParam.SetTextRightOffset:=True;
       CBParam.TextRightOffset:=StrToIntDef(Value,0);
      end else
     if S='TextBottomOffset' then
      begin
       CBParam.SetTextBottomOffset:=True;
       CBParam.TextBottomOffset:=StrToIntDef(Value,0);
      end else
       TryToSetFontDataParameter(S,Value,CBParam.FD);
   end;
end;

procedure TSXSkinCheckBoxStyle.GetCurrentCBState(var CBState:TSXSkinCheckBoxStateParam;
           State:TCheckBoxState;Down,MouseOver,Focused,Enabled:Boolean);

 procedure SetCBStateFrom(T:TSXSkinCheckBoxStateParam);
 begin
  if CBState.Style='' then CBState.Style:=T.Style;
  if CBState.OverStyle='' then CBState.OverStyle:=T.OverStyle;
  if not CBState.SetCaptionLeftOffset and T.SetCaptionLeftOffset then
   begin
    CBState.SetCaptionLeftOffset:=True;
    CBState.CaptionLeftOffset:=T.CaptionLeftOffset;
   end;
  if not CBState.SetCaptionTopOffset and T.SetCaptionTopOffset then
   begin
    CBState.SetCaptionTopOffset:=True;
    CBState.CaptionTopOffset:=T.CaptionTopOffset;
   end;
  if not CBState.SetCaptionRightOffset and T.SetCaptionRightOffset then
   begin
    CBState.SetCaptionRightOffset:=True;
    CBState.CaptionRightOffset:=T.CaptionRightOffset;
   end;
  if not CBState.SetCaptionBottomOffset and T.SetCaptionBottomOffset then
   begin
    CBState.SetCaptionBottomOffset:=True;
    CBState.CaptionBottomOffset:=T.CaptionBottomOffset;
   end;
  if not CBState.SetTextLeftOffset and T.SetTextLeftOffset then
   begin
    CBState.SetTextLeftOffset:=True;
    CBState.TextLeftOffset:=T.TextLeftOffset;
   end;
  if not CBState.SetTextTopOffset and T.SetTextTopOffset then
   begin
    CBState.SetTextTopOffset:=True;
    CBState.TextTopOffset:=T.TextTopOffset;
   end;
  if not CBState.SetTextRightOffset and T.SetTextRightOffset then
   begin
    CBState.SetTextRightOffset:=True;
    CBState.TextRightOffset:=T.TextRightOffset;
   end;
  if not CBState.SetTextBottomOffset and T.SetTextBottomOffset then
   begin
    CBState.SetTextBottomOffset:=True;
    CBState.TextBottomOffset:=T.TextBottomOffset;
   end;
  if CBState.GlyphStyle='' then CBState.GlyphStyle:=T.GlyphStyle;
  if CBState.GlyphWidth=0 then CBState.GlyphWidth:=T.GlyphWidth;
  if CBState.GlyphHeight=0 then CBState.GlyphHeight:=T.GlyphHeight;
  AddFontData(CBState.FD,T.FD);
 end;

begin
 Finalize(CBState);
 FillChar(CBState,sizeof(CBState),0);
 case State of
  cbChecked:   begin
                if not Enabled then
                 begin
                  SetCBStateFrom(RUCState);
                  SetCBStateFrom(NUCState);
                  SetCBStateFrom(RUUState);
                  SetCBStateFrom(NUUState);
                 end else
                if Down then
                 begin
                  SetCBStateFrom(DFCState);
                  SetCBStateFrom(DFUState);
                  SetCBStateFrom(NFCState);
                  SetCBStateFrom(NFUState);
                  SetCBStateFrom(NUCState);
                  SetCBStateFrom(NUUState);
                 end else
                if MouseOver then
                 begin
                  if Focused then
                   begin
                    SetCBStateFrom(HFCState);
                    SetCBStateFrom(HUCState);
                    SetCBStateFrom(HFUState);
                    SetCBStateFrom(HUUState);
                    SetCBStateFrom(NFCState);
                    SetCBStateFrom(NUCState);
                    SetCBStateFrom(NFUState);
                    SetCBStateFrom(NUUState);
                   end else
                    begin
                     SetCBStateFrom(HUCState);
                     SetCBStateFrom(NUCState);
                     SetCBStateFrom(HUUState);
                     SetCBStateFrom(NUUState);
                    end;
                 end else
                  begin
                   if Focused then
                    begin
                     SetCBStateFrom(NFCState);
                     SetCBStateFrom(NUCState);
                     SetCBStateFrom(NFUState);
                     SetCBStateFrom(NUUState);
                    end else
                     begin
                      SetCBStateFrom(NUCState);
                      SetCBStateFrom(NUUState);
                     end;
                  end;
               end;
  cbGrayed:    begin
                if not Enabled then
                 begin
                  SetCBStateFrom(RUGState);
                  SetCBStateFrom(NUGState);
                  SetCBStateFrom(RUCState);
                  SetCBStateFrom(NUCState);
                  SetCBStateFrom(RUUState);
                  SetCBStateFrom(NUUState);
                 end else
                if Down then
                 begin
                  SetCBStateFrom(DFGState);
                  SetCBStateFrom(DFCState);
                  SetCBStateFrom(DFUState);
                  SetCBStateFrom(NFGState);
                  SetCBStateFrom(NFCState);
                  SetCBStateFrom(NFUState);
                  SetCBStateFrom(NUGState);
                  SetCBStateFrom(NUCState);
                  SetCBStateFrom(NUUState);
                 end else
                if MouseOver then
                 begin
                  if Focused then
                   begin
                    SetCBStateFrom(HFGState);
                    SetCBStateFrom(HUGState);
                    SetCBStateFrom(HFCState);
                    SetCBStateFrom(HUCState);
                    SetCBStateFrom(HFUState);
                    SetCBStateFrom(HUUState);
                    SetCBStateFrom(NFGState);
                    SetCBStateFrom(NUGState);
                    SetCBStateFrom(NFCState);
                    SetCBStateFrom(NUCState);
                    SetCBStateFrom(NFUState);
                    SetCBStateFrom(NUUState);
                   end else
                    begin
                     SetCBStateFrom(HUGState);
                     SetCBStateFrom(NUGState);
                     SetCBStateFrom(HUCState);
                     SetCBStateFrom(NUCState);
                     SetCBStateFrom(HUUState);
                     SetCBStateFrom(NUUState);
                    end;
                 end else
                  begin
                   if Focused then
                    begin
                     SetCBStateFrom(NFGState);
                     SetCBStateFrom(NUGState);
                     SetCBStateFrom(NFCState);
                     SetCBStateFrom(NUCState);
                     SetCBStateFrom(NFUState);
                     SetCBStateFrom(NUUState);
                    end else
                     begin
                      SetCBStateFrom(NUGState);
                      SetCBStateFrom(NUCState);
                      SetCBStateFrom(NUUState);
                     end;
                  end;
               end;
  cbUnchecked: begin
                if not Enabled then
                 begin
                  SetCBStateFrom(RUUState);
                  SetCBStateFrom(NUUState);
                 end else
                if Down then
                 begin
                  SetCBStateFrom(DFUState);
                  SetCBStateFrom(NFUState);
                  SetCBStateFrom(NUUState);
                 end else
                if MouseOver then
                 begin
                  if Focused then
                   begin
                    SetCBStateFrom(HFUState);
                    SetCBStateFrom(HUUState);
                    SetCBStateFrom(NFUState);
                    SetCBStateFrom(NUUState);
                   end else
                    begin
                     SetCBStateFrom(HUUState);
                     SetCBStateFrom(NUUState);
                    end;
                 end else
                  begin
                   if Focused then
                    begin
                     SetCBStateFrom(NFUState);
                     SetCBStateFrom(NUUState);
                    end else
                     begin
                      SetCBStateFrom(NUUState);
                     end;
                  end;
               end;
 end;              
end;

{ TSXSkinRadioButtonStyle }

procedure TSXSkinRadioButtonStyle.AssignStyle(Style:TSXSkinStyle);
begin
 inherited;
 if Style is TSXSkinRadioButtonStyle then
  begin
   NUUState:=TSXSkinRadioButtonStyle(Style).NUUState;
   NFUState:=TSXSkinRadioButtonStyle(Style).NFUState;
   HUUState:=TSXSkinRadioButtonStyle(Style).HUUState;
   HFUState:=TSXSkinRadioButtonStyle(Style).HFUState;
   DFUState:=TSXSkinRadioButtonStyle(Style).DFUState;
   RUUState:=TSXSkinRadioButtonStyle(Style).RUUState;
   NUCState:=TSXSkinRadioButtonStyle(Style).NUCState;
   NFCState:=TSXSkinRadioButtonStyle(Style).NFCState;
   HUCState:=TSXSkinRadioButtonStyle(Style).HUCState;
   HFCState:=TSXSkinRadioButtonStyle(Style).HFCState;
   DFCState:=TSXSkinRadioButtonStyle(Style).DFCState;
   RUCState:=TSXSkinRadioButtonStyle(Style).RUCState;
   //
   HInRadioButtonEffect:=TSXSkinRadioButtonStyle(Style).HInRadioButtonEffect;
   HOutRadioButtonEffect:=TSXSkinRadioButtonStyle(Style).HOutRadioButtonEffect;
   CheckRadioButtonEffect:=TSXSkinRadioButtonStyle(Style).CheckRadioButtonEffect;
   UncheckRadioButtonEffect:=TSXSkinRadioButtonStyle(Style).UncheckRadioButtonEffect;
   DownRadioButtonEffect:=TSXSkinRadioButtonStyle(Style).DownRadioButtonEffect;
   UpRadioButtonEffect:=TSXSkinRadioButtonStyle(Style).UpRadioButtonEffect;
   EnableRadioButtonEffect:=TSXSkinRadioButtonStyle(Style).EnableRadioButtonEffect;
   DisableRadioButtonEffect:=TSXSkinRadioButtonStyle(Style).DisableRadioButtonEffect;
   FocusRadioButtonEffect:=TSXSkinRadioButtonStyle(Style).FocusRadioButtonEffect;
   UnfocusRadioButtonEffect:=TSXSkinRadioButtonStyle(Style).UnfocusRadioButtonEffect;
   //
   HInGlyphEffect:=TSXSkinButtonStyle(Style).HInGlyphEffect;
   HOutGlyphEffect:=TSXSkinButtonStyle(Style).HOutGlyphEffect;
   CheckGlyphEffect:=TSXSkinButtonStyle(Style).CheckGlyphEffect;
   UncheckGlyphEffect:=TSXSkinButtonStyle(Style).UncheckGlyphEffect;
   DownGlyphEffect:=TSXSkinButtonStyle(Style).DownGlyphEffect;
   UpGlyphEffect:=TSXSkinButtonStyle(Style).UpGlyphEffect;
   EnableGlyphEffect:=TSXSkinButtonStyle(Style).EnableGlyphEffect;
   DisableGlyphEffect:=TSXSkinButtonStyle(Style).DisableGlyphEffect;
   FocusGlyphEffect:=TSXSkinButtonStyle(Style).FocusGlyphEffect;
   UnfocusGlyphEffect:=TSXSkinButtonStyle(Style).UnfocusGlyphEffect;
  end;
end;

procedure TSXSkinRadioButtonStyle.SetParameter(const Name,Value:String);
var RBParam:PSXSkinRadioButtonStateParam;
          S:String;
begin
 RBParam:=nil;
 S:=Copy(Name,1,4);
 if S='NUU_' then RBParam:=@NUUState else
  if S='NFU_' then RBParam:=@NFUState else
  if S='HUU_' then RBParam:=@HUUState else
  if S='HFU_' then RBParam:=@HFUState else
  if S='DFU_' then RBParam:=@DFUState else
  if S='RUU_' then RBParam:=@RUUState else
  if S='NUC_' then RBParam:=@NUCState else
  if S='NFC_' then RBParam:=@NFCState else
  if S='HUC_' then RBParam:=@HUCState else
  if S='HFC_' then RBParam:=@HFCState else
  if S='DFC_' then RBParam:=@DFCState else
  if S='RUC_' then RBParam:=@RUCState;
 if RBParam=nil then
  begin
   if not TryToSetTransformEffectDataParameter(Name,Value,HInRadioButtonEffect,'HighlightInRadioButton') then
    if not TryToSetTransformEffectDataParameter(Name,Value,HOutRadioButtonEffect,'HighlightOutRadioButton') then
    if not TryToSetTransformEffectDataParameter(Name,Value,CheckRadioButtonEffect,'CheckRadioButton') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UncheckRadioButtonEffect,'UncheckRadioButton') then
    if not TryToSetTransformEffectDataParameter(Name,Value,DownRadioButtonEffect,'DownRadioButton') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UpRadioButtonEffect,'UpRadioButton') then
    if not TryToSetTransformEffectDataParameter(Name,Value,EnableRadioButtonEffect,'EnableRadioButton') then
    if not TryToSetTransformEffectDataParameter(Name,Value,DisableRadioButtonEffect,'DisableRadioButton') then
    if not TryToSetTransformEffectDataParameter(Name,Value,FocusRadioButtonEffect,'FocusRadioButton') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UnfocusRadioButtonEffect,'UnfocusRadioButton') then
    //
    if not TryToSetTransformEffectDataParameter(Name,Value,HInGlyphEffect,'HighlightInGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,HOutGlyphEffect,'HighlightOutGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,CheckGlyphEffect,'CheckGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UncheckGlyphEffect,'UncheckGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,DownGlyphEffect,'DownGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UpGlyphEffect,'UpGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,EnableGlyphEffect,'EnableGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,DisableGlyphEffect,'DisableGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,FocusGlyphEffect,'FocusGlyph') then
     TryToSetTransformEffectDataParameter(Name,Value,UnfocusGlyphEffect,'UnfocusGlyph');
  end else
   begin
    S:=Copy(Name,5,MaxInt);
    if S='Style' then RBParam.Style:=Value else
     if S='OverStyle' then RBParam.OverStyle:=Value else
     if S='Glyph' then RBParam.GlyphStyle:=Value else
     if S='GlyphWidth' then RBParam.GlyphWidth:=StrToIntDef(Value,0) else
     if S='GlyphHeight' then RBParam.GlyphHeight:=StrToIntDef(Value,0) else
     if S='CaptionLeftOffset' then
      begin
       RBParam.SetCaptionLeftOffset:=True;
       RBParam.CaptionLeftOffset:=StrToIntDef(Value,0);
      end else
     if S='CaptionTopOffset' then
      begin
       RBParam.SetCaptionTopOffset:=True;
       RBParam.CaptionTopOffset:=StrToIntDef(Value,0);
      end else
     if S='CaptionRightOffset' then
      begin
       RBParam.SetCaptionRightOffset:=True;
       RBParam.CaptionRightOffset:=StrToIntDef(Value,0);
      end else
     if S='CaptionBottomOffset' then
      begin
       RBParam.SetCaptionBottomOffset:=True;
       RBParam.CaptionBottomOffset:=StrToIntDef(Value,0);
      end else
     if S='TextLeftOffset' then
      begin
       RBParam.SetTextLeftOffset:=True;
       RBParam.TextLeftOffset:=StrToIntDef(Value,0);
      end else
     if S='TextTopOffset' then
      begin
       RBParam.SetTextTopOffset:=True;
       RBParam.TextTopOffset:=StrToIntDef(Value,0);
      end else
     if S='TextRightOffset' then
      begin
       RBParam.SetTextRightOffset:=True;
       RBParam.TextRightOffset:=StrToIntDef(Value,0);
      end else
     if S='TextBottomOffset' then
      begin
       RBParam.SetTextBottomOffset:=True;
       RBParam.TextBottomOffset:=StrToIntDef(Value,0);
      end else
       TryToSetFontDataParameter(S,Value,RBParam.FD);
   end;
end;

procedure TSXSkinRadioButtonStyle.GetCurrentRBState(var RBState:TSXSkinRadioButtonStateParam;
           Checked,Down,MouseOver,Focused,Enabled:Boolean);

 procedure SetRBStateFrom(T:TSXSkinRadioButtonStateParam);
 begin
  if RBState.Style='' then RBState.Style:=T.Style;
  if RBState.OverStyle='' then RBState.OverStyle:=T.OverStyle;
  if not RBState.SetCaptionLeftOffset and T.SetCaptionLeftOffset then
   begin
    RBState.SetCaptionLeftOffset:=True;
    RBState.CaptionLeftOffset:=T.CaptionLeftOffset;
   end;
  if not RBState.SetCaptionTopOffset and T.SetCaptionTopOffset then
   begin
    RBState.SetCaptionTopOffset:=True;
    RBState.CaptionTopOffset:=T.CaptionTopOffset;
   end;
  if not RBState.SetCaptionRightOffset and T.SetCaptionRightOffset then
   begin
    RBState.SetCaptionRightOffset:=True;
    RBState.CaptionRightOffset:=T.CaptionRightOffset;
   end;
  if not RBState.SetCaptionBottomOffset and T.SetCaptionBottomOffset then
   begin
    RBState.SetCaptionBottomOffset:=True;
    RBState.CaptionBottomOffset:=T.CaptionBottomOffset;
   end;
  if not RBState.SetTextLeftOffset and T.SetTextLeftOffset then
   begin
    RBState.SetTextLeftOffset:=True;
    RBState.TextLeftOffset:=T.TextLeftOffset;
   end;
  if not RBState.SetTextTopOffset and T.SetTextTopOffset then
   begin
    RBState.SetTextTopOffset:=True;
    RBState.TextTopOffset:=T.TextTopOffset;
   end;
  if not RBState.SetTextRightOffset and T.SetTextRightOffset then
   begin
    RBState.SetTextRightOffset:=True;
    RBState.TextRightOffset:=T.TextRightOffset;
   end;
  if not RBState.SetTextBottomOffset and T.SetTextBottomOffset then
   begin
    RBState.SetTextBottomOffset:=True;
    RBState.TextBottomOffset:=T.TextBottomOffset;
   end;
  if RBState.GlyphStyle='' then RBState.GlyphStyle:=T.GlyphStyle;
  if RBState.GlyphWidth=0 then RBState.GlyphWidth:=T.GlyphWidth;
  if RBState.GlyphHeight=0 then RBState.GlyphHeight:=T.GlyphHeight;
  AddFontData(RBState.FD,T.FD);
 end;

begin
 Finalize(RBState);
 FillChar(RBState,sizeof(RBState),0);
 if Checked then
  begin
   if not Enabled then
    begin
     SetRBStateFrom(RUCState);
     SetRBStateFrom(NUCState);
     SetRBStateFrom(RUUState);
     SetRBStateFrom(NUUState);
    end else
   if Down then
    begin
     SetRBStateFrom(DFCState);
     SetRBStateFrom(DFUState);
     SetRBStateFrom(NFCState);
     SetRBStateFrom(NFUState);
     SetRBStateFrom(NUCState);
     SetRBStateFrom(NUUState);
    end else
   if MouseOver then
    begin
     if Focused then
      begin
       SetRBStateFrom(HFCState);
       SetRBStateFrom(HUCState);
       SetRBStateFrom(HFUState);
       SetRBStateFrom(HUUState);
       SetRBStateFrom(NFCState);
       SetRBStateFrom(NUCState);
       SetRBStateFrom(NFUState);
       SetRBStateFrom(NUUState);
      end else
       begin
        SetRBStateFrom(HUCState);
        SetRBStateFrom(NUCState);
        SetRBStateFrom(HUUState);
        SetRBStateFrom(NUUState);
       end;
    end else
     begin
      if Focused then
       begin
        SetRBStateFrom(NFCState);
        SetRBStateFrom(NUCState);
        SetRBStateFrom(NFUState);
        SetRBStateFrom(NUUState);
       end else
        begin
         SetRBStateFrom(NUCState);
         SetRBStateFrom(NUUState);
        end;
     end;
  end else
   begin
    if not Enabled then
     begin
      SetRBStateFrom(RUUState);
      SetRBStateFrom(NUUState);
     end else
    if Down then
     begin
      SetRBStateFrom(DFUState);
      SetRBStateFrom(NFUState);
      SetRBStateFrom(NUUState);
     end else
    if MouseOver then
     begin
      if Focused then
       begin
        SetRBStateFrom(HFUState);
        SetRBStateFrom(HUUState);
        SetRBStateFrom(NFUState);
        SetRBStateFrom(NUUState);
       end else
        begin
         SetRBStateFrom(HUUState);
         SetRBStateFrom(NUUState);
        end;
     end else
      begin
       if Focused then
        begin
         SetRBStateFrom(NFUState);
         SetRBStateFrom(NUUState);
        end else
         begin
          SetRBStateFrom(NUUState);
         end;
      end;
   end;
end;

{ TSXSkinGroupBoxStyle }

procedure TSXSkinGroupBoxStyle.AssignStyle(Style:TSXSkinStyle);
begin
 inherited;
 if Style is TSXSkinGroupBoxStyle then
  begin
   LabelStyle:=TSXSkinGroupBoxStyle(Style).LabelStyle;
   NUState:=TSXSkinGroupBoxStyle(Style).NUState;
   NFState:=TSXSkinGroupBoxStyle(Style).NFState;
   HUState:=TSXSkinGroupBoxStyle(Style).HUState;
   HFState:=TSXSkinGroupBoxStyle(Style).HFState;
   RUState:=TSXSkinGroupBoxStyle(Style).RUState;
   GlyphWidth:=TSXSkinGroupBoxStyle(Style).GlyphWidth;
   GlyphHeight:=TSXSkinGroupBoxStyle(Style).GlyphHeight;
   ClientRect:=TSXSkinGroupBoxStyle(Style).ClientRect;
  end;
end;

procedure TSXSkinGroupBoxStyle.SetParameter(const Name,Value:String);
var GBParam:PSXSkinGroupBoxStateParam;
          S:String;
begin
 GBParam:=nil;
 S:=Copy(Name,1,3);
 if S='NU_' then GBParam:=@NUState else
  if S='NF_' then GBParam:=@NFState else
  if S='HU_' then GBParam:=@HUState else
  if S='HF_' then GBParam:=@HFState else
  if S='RU_' then GBParam:=@RUState;
 if GBParam=nil then
  begin
   if Name='LabelStyle' then
    LabelStyle:=Value else
   if Name='GlyphWidth' then
    GlyphWidth:=StrToIntDef(Value,0) else
   if Name='GlyphHeight' then
    GlyphHeight:=StrToIntDef(Value,0) else
   if Name='ClientRect' then
    ClientRect:=WithoutAllSpaces(Value);
  end else
   begin
    S:=Copy(Name,4,MaxInt);
    if S='Style' then GBParam.Style:=Value else
     if S='OverStyle' then GBParam.OverStyle:=Value else
     if S='CaptionPosition' then
      begin
       GBParam.SetCaptionPosition:=True;
       GBParam.CaptionPosition:=StrToIntDef(Value,0);
      end else
     if S='CaptionLeftOffset' then
      begin
       GBParam.SetCaptionLeftOffset:=True;
       GBParam.CaptionLeftOffset:=StrToIntDef(Value,0);
      end else
     if S='CaptionTopOffset' then
      begin
       GBParam.SetCaptionTopOffset:=True;
       GBParam.CaptionTopOffset:=StrToIntDef(Value,0);
      end else
     if S='CaptionRightOffset' then
      begin
       GBParam.SetCaptionRightOffset:=True;
       GBParam.CaptionRightOffset:=StrToIntDef(Value,0);
      end else
     if S='CaptionBottomOffset' then
      begin
       GBParam.SetCaptionBottomOffset:=True;
       GBParam.CaptionBottomOffset:=StrToIntDef(Value,0);
      end else
     if S='TextLeftOffset' then
      begin
       GBParam.SetTextLeftOffset:=True;
       GBParam.TextLeftOffset:=StrToIntDef(Value,0);
      end else
     if S='TextTopOffset' then
      begin
       GBParam.SetTextTopOffset:=True;
       GBParam.TextTopOffset:=StrToIntDef(Value,0);
      end else
     if S='TextRightOffset' then
      begin
       GBParam.SetTextRightOffset:=True;
       GBParam.TextRightOffset:=StrToIntDef(Value,0);
      end else
     if S='TextBottomOffset' then
      begin
       GBParam.SetTextBottomOffset:=True;
       GBParam.TextBottomOffset:=StrToIntDef(Value,0);
      end else
     if S='TransparentRect' then
      GBParam.TransparentRect:=WithoutAllSpaces(Value);
    end;
end;

procedure TSXSkinGroupBoxStyle.GetCurrentGBState(var GBState:TSXSkinGroupBoxStateParam;
           MouseOver,Focused,Enabled:Boolean);

 procedure SetGBStateFrom(T:TSXSkinGroupBoxStateParam);
 begin
  if GBState.Style='' then GBState.Style:=T.Style;
  if GBState.OverStyle='' then GBState.OverStyle:=T.OverStyle;
  if not GBState.SetCaptionPosition and T.SetCaptionPosition then
   begin
    GBState.SetCaptionPosition:=True;
    GBState.CaptionPosition:=T.CaptionPosition;
   end;
  if not GBState.SetCaptionLeftOffset and T.SetCaptionLeftOffset then
   begin
    GBState.SetCaptionLeftOffset:=True;
    GBState.CaptionLeftOffset:=T.CaptionLeftOffset;
   end;
  if not GBState.SetCaptionTopOffset and T.SetCaptionTopOffset then
   begin
    GBState.SetCaptionTopOffset:=True;
    GBState.CaptionTopOffset:=T.CaptionTopOffset;
   end;
  if not GBState.SetCaptionRightOffset and T.SetCaptionRightOffset then
   begin
    GBState.SetCaptionRightOffset:=True;
    GBState.CaptionRightOffset:=T.CaptionRightOffset;
   end;
  if not GBState.SetCaptionBottomOffset and T.SetCaptionBottomOffset then
   begin
    GBState.SetCaptionBottomOffset:=True;
    GBState.CaptionBottomOffset:=T.CaptionBottomOffset;
   end;
  if not GBState.SetTextLeftOffset and T.SetTextLeftOffset then
   begin
    GBState.SetTextLeftOffset:=True;
    GBState.TextLeftOffset:=T.TextLeftOffset;
   end;
  if not GBState.SetTextTopOffset and T.SetTextTopOffset then
   begin
    GBState.SetTextTopOffset:=True;
    GBState.TextTopOffset:=T.TextTopOffset;
   end;
  if not GBState.SetTextRightOffset and T.SetTextRightOffset then
   begin
    GBState.SetTextRightOffset:=True;
    GBState.TextRightOffset:=T.TextRightOffset;
   end;
  if not GBState.SetTextBottomOffset and T.SetTextBottomOffset then
   begin
    GBState.SetTextBottomOffset:=True;
    GBState.TextBottomOffset:=T.TextBottomOffset;
   end;
  if GBState.TransparentRect='' then GBState.TransparentRect:=T.TransparentRect;
 end;

begin
 Finalize(GBState);
 FillChar(GBState,sizeof(GBState),0);
 if not Enabled then
  begin
   SetGBStateFrom(RUState);
   SetGBStateFrom(NUState);
  end else
 if MouseOver then
  begin
   if Focused then
    begin
     SetGBStateFrom(HFState);
     SetGBStateFrom(HUState);
     SetGBStateFrom(NFState);
     SetGBStateFrom(NUState);
    end else
     begin
      SetGBStateFrom(HUState);
      SetGBStateFrom(NUState);
     end;
  end else
   begin
    if Focused then
     begin
      SetGBStateFrom(NFState);
      SetGBStateFrom(NUState);
     end else
      begin
       SetGBStateFrom(NUState);
      end;
   end;
end;

{ TSXSkinMultiStateStyle }

procedure TSXSkinMultiStateStyle.AssignStyle(Style:TSXSkinStyle);
begin
 inherited;
 if Style is TSXSkinMultiStateStyle then
  begin
   NStyle:=TSXSkinMultiStateStyle(Style).NStyle;
   HStyle:=TSXSkinMultiStateStyle(Style).HStyle;
   DStyle:=TSXSkinMultiStateStyle(Style).DStyle;
   RStyle:=TSXSkinMultiStateStyle(Style).RStyle;
  end;
end;

procedure TSXSkinMultiStateStyle.SetParameter(const Name,Value:String);
begin
 if Name='NStyle' then NStyle:=Value else
 if Name='HStyle' then HStyle:=Value else
 if Name='DStyle' then DStyle:=Value else
 if Name='RStyle' then RStyle:=Value;
end;

{ TSXSkinMultiStateCheckStyle }

procedure TSXSkinMultiStateCheckStyle.AssignStyle(Style:TSXSkinStyle);
begin
 inherited;
 if Style is TSXSkinMultiStateCheckStyle then
  begin
   NUStyle:=TSXSkinMultiStateCheckStyle(Style).NUStyle;
   HUStyle:=TSXSkinMultiStateCheckStyle(Style).HUStyle;
   DUStyle:=TSXSkinMultiStateCheckStyle(Style).DUStyle;
   RUStyle:=TSXSkinMultiStateCheckStyle(Style).RUStyle;
   NCStyle:=TSXSkinMultiStateCheckStyle(Style).NCStyle;
   HCStyle:=TSXSkinMultiStateCheckStyle(Style).HCStyle;
   DCStyle:=TSXSkinMultiStateCheckStyle(Style).DCStyle;
   RCStyle:=TSXSkinMultiStateCheckStyle(Style).RCStyle;
  end;
end;

procedure TSXSkinMultiStateCheckStyle.SetParameter(const Name,Value:String);
begin
 if Name='NUStyle' then NUStyle:=Value else
 if Name='HUStyle' then HUStyle:=Value else
 if Name='DUStyle' then DUStyle:=Value else
 if Name='RUStyle' then RUStyle:=Value else
 if Name='NCStyle' then NCStyle:=Value else
 if Name='HCStyle' then HCStyle:=Value else
 if Name='DCStyle' then DCStyle:=Value else
 if Name='RCStyle' then RCStyle:=Value;
end;

{ TSXSkinLabelStyle }

procedure TSXSkinLabelStyle.AssignStyle(Style:TSXSkinStyle);
begin
 inherited;
 if Style is TSXSkinLabelStyle then
  begin
   NState:=TSXSkinLabelStyle(Style).NState;
   HState:=TSXSkinLabelStyle(Style).HState;
   RState:=TSXSkinLabelStyle(Style).RState;
  end;
end;

procedure TSXSkinLabelStyle.SetParameter(const Name,Value:String);
var LParam:PSXSkinLabelStateParam;
         S:String;
begin
 LParam:=nil;
 S:=Copy(Name,1,2);
 if S='N_' then LParam:=@NState else
  if S='H_' then LParam:=@HState else
  if S='R_' then LParam:=@RState;
 if LParam=nil then
  begin
   //
  end else
   begin
    S:=Copy(Name,3,MaxInt);
    TryToSetFontDataParameter(S,Value,LParam.FD);
   end;
end;

{ TSXSkinButtonStyle }

procedure TSXSkinButtonStyle.AssignStyle(Style:TSXSkinStyle);
begin
 inherited;
 if Style is TSXSkinButtonStyle then
  begin
   NUUState:=TSXSkinButtonStyle(Style).NUUState;
   NFUState:=TSXSkinButtonStyle(Style).NFUState;
   HUUState:=TSXSkinButtonStyle(Style).HUUState;
   HFUState:=TSXSkinButtonStyle(Style).HFUState;
   DFUState:=TSXSkinButtonStyle(Style).DFUState;
   RUUState:=TSXSkinButtonStyle(Style).RUUState;
   NUCState:=TSXSkinButtonStyle(Style).NUCState;
   NFCState:=TSXSkinButtonStyle(Style).NFCState;
   HUCState:=TSXSkinButtonStyle(Style).HUCState;
   HFCState:=TSXSkinButtonStyle(Style).HFCState;
   DFCState:=TSXSkinButtonStyle(Style).DFCState;
   RUCState:=TSXSkinButtonStyle(Style).RUCState;
   //
   HInButtonEffect:=TSXSkinButtonStyle(Style).HInButtonEffect;
   HOutButtonEffect:=TSXSkinButtonStyle(Style).HOutButtonEffect;
   CheckButtonEffect:=TSXSkinButtonStyle(Style).CheckButtonEffect;
   UncheckButtonEffect:=TSXSkinButtonStyle(Style).UncheckButtonEffect;
   DownButtonEffect:=TSXSkinButtonStyle(Style).DownButtonEffect;
   UpButtonEffect:=TSXSkinButtonStyle(Style).UpButtonEffect;
   EnableButtonEffect:=TSXSkinButtonStyle(Style).EnableButtonEffect;
   DisableButtonEffect:=TSXSkinButtonStyle(Style).DisableButtonEffect;
   FocusButtonEffect:=TSXSkinButtonStyle(Style).FocusButtonEffect;
   UnfocusButtonEffect:=TSXSkinButtonStyle(Style).UnfocusButtonEffect;
   //
   HInGlyphEffect:=TSXSkinButtonStyle(Style).HInGlyphEffect;
   HOutGlyphEffect:=TSXSkinButtonStyle(Style).HOutGlyphEffect;
   CheckGlyphEffect:=TSXSkinButtonStyle(Style).CheckGlyphEffect;
   UncheckGlyphEffect:=TSXSkinButtonStyle(Style).UncheckGlyphEffect;
   DownGlyphEffect:=TSXSkinButtonStyle(Style).DownGlyphEffect;
   UpGlyphEffect:=TSXSkinButtonStyle(Style).UpGlyphEffect;
   EnableGlyphEffect:=TSXSkinButtonStyle(Style).EnableGlyphEffect;
   DisableGlyphEffect:=TSXSkinButtonStyle(Style).DisableGlyphEffect;
   FocusGlyphEffect:=TSXSkinButtonStyle(Style).FocusGlyphEffect;
   UnfocusGlyphEffect:=TSXSkinButtonStyle(Style).UnfocusGlyphEffect;
   //
   HInDDGlyphEffect:=TSXSkinButtonStyle(Style).HInDDGlyphEffect;
   HOutDDGlyphEffect:=TSXSkinButtonStyle(Style).HOutDDGlyphEffect;
   CheckDDGlyphEffect:=TSXSkinButtonStyle(Style).CheckDDGlyphEffect;
   UncheckDDGlyphEffect:=TSXSkinButtonStyle(Style).UncheckDDGlyphEffect;
   DownDDGlyphEffect:=TSXSkinButtonStyle(Style).DownDDGlyphEffect;
   UpDDGlyphEffect:=TSXSkinButtonStyle(Style).UpDDGlyphEffect;
   EnableDDGlyphEffect:=TSXSkinButtonStyle(Style).EnableDDGlyphEffect;
   DisableDDGlyphEffect:=TSXSkinButtonStyle(Style).DisableDDGlyphEffect;
   FocusDDGlyphEffect:=TSXSkinButtonStyle(Style).FocusDDGlyphEffect;
   UnfocusDDGlyphEffect:=TSXSkinButtonStyle(Style).UnfocusDDGlyphEffect;
   //
   DropDownGlyphWidth:=TSXSkinButtonStyle(Style).DropDownGlyphWidth;
   DropDownGlyphHeight:=TSXSkinButtonStyle(Style).DropDownGlyphHeight;
   DropDownGlyphOffset:=TSXSkinButtonStyle(Style).DropDownGlyphOffset;
  end;
end;

procedure TSXSkinButtonStyle.SetParameter(const Name,Value:String);
var BParam:PSXSkinButtonStateParam;
      S,S2:String;
begin
 BParam:=nil;
 S:=Copy(Name,1,3);
 S2:=Copy(Name,1,4);
 if (S='NU_') or (S2='NUU_') then BParam:=@NUUState else
  if (S='NF_') or (S2='NFU_') then BParam:=@NFUState else
  if (S='HU_') or (S2='HUU_') then BParam:=@HUUState else
  if (S='HF_') or (S2='HFU_') then BParam:=@HFUState else
  if (S='DF_') or (S2='DFU_') then BParam:=@DFUState else
  if (S='RU_') or (S2='RUU_') then BParam:=@RUUState else
  if S2='NUC_' then BParam:=@NUCState else
  if S2='NFC_' then BParam:=@NFCState else
  if S2='HUC_' then BParam:=@HUCState else
  if S2='HFC_' then BParam:=@HFCState else
  if S2='DFC_' then BParam:=@DFCState else
  if S2='RUC_' then BParam:=@RUCState;
 if BParam=nil then
  begin
   if not TryToSetTransformEffectDataParameter(Name,Value,HInButtonEffect,'HighlightInButton') then
    if not TryToSetTransformEffectDataParameter(Name,Value,HOutButtonEffect,'HighlightOutButton') then
    if not TryToSetTransformEffectDataParameter(Name,Value,CheckButtonEffect,'CheckButton') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UncheckButtonEffect,'UncheckButton') then
    if not TryToSetTransformEffectDataParameter(Name,Value,DownButtonEffect,'DownButton') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UpButtonEffect,'UpButton') then
    if not TryToSetTransformEffectDataParameter(Name,Value,EnableButtonEffect,'EnableButton') then
    if not TryToSetTransformEffectDataParameter(Name,Value,DisableButtonEffect,'DisableButton') then
    if not TryToSetTransformEffectDataParameter(Name,Value,FocusButtonEffect,'FocusButton') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UnfocusButtonEffect,'UnfocusButton') then
    //
    if not TryToSetTransformEffectDataParameter(Name,Value,HInGlyphEffect,'HighlightInGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,HOutGlyphEffect,'HighlightOutGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,CheckGlyphEffect,'CheckGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UncheckGlyphEffect,'UncheckGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,DownGlyphEffect,'DownGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UpGlyphEffect,'UpGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,EnableGlyphEffect,'EnableGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,DisableGlyphEffect,'DisableGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,FocusGlyphEffect,'FocusGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UnfocusGlyphEffect,'UnfocusGlyph') then
    //
    if not TryToSetTransformEffectDataParameter(Name,Value,HInDDGlyphEffect,'HighlightInDDGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,HOutDDGlyphEffect,'HighlightOutDDGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,CheckDDGlyphEffect,'CheckDDGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UncheckDDGlyphEffect,'UncheckDDGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,DownDDGlyphEffect,'DownDDGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UpDDGlyphEffect,'UpDDGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,EnableDDGlyphEffect,'EnableDDGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,DisableDDGlyphEffect,'DisableDDGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,FocusDDGlyphEffect,'FocusDDGlyph') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UnfocusDDGlyphEffect,'UnfocusDDGlyph') then
    //
    if Name='DropDownGlyphWidth' then
     DropDownGlyphWidth:=StrToIntDef(Value,0) else
    if Name='DropDownGlyphHeight' then
     DropDownGlyphHeight:=StrToIntDef(Value,0) else
    if Name='DropDownGlyphOffset' then
     DropDownGlyphOffset:=StrToIntDef(Value,0);
  end else
   begin
    if Name[3]='_' then S:=Copy(Name,4,MaxInt) else
     S:=Copy(Name,5,MaxInt);
    if S='Style' then BParam.Style:=Value else
    if S='OverStyle' then BParam.OverStyle:=Value else
    if S='CaptionLeftOffset' then
     begin
      BParam.SetCaptionLeftOffset:=True;
      BParam.CaptionLeftOffset:=StrToIntDef(Value,0);
     end else
    if S='CaptionTopOffset' then
     begin
      BParam.SetCaptionTopOffset:=True;
      BParam.CaptionTopOffset:=StrToIntDef(Value,0);
     end else
    if S='CaptionRightOffset' then
     begin
      BParam.SetCaptionRightOffset:=True;
      BParam.CaptionRightOffset:=StrToIntDef(Value,0);
     end else
    if S='CaptionBottomOffset' then
     begin
      BParam.SetCaptionBottomOffset:=True;
      BParam.CaptionBottomOffset:=StrToIntDef(Value,0);
     end else
    if not TryToSetFilterDataParameter(S,Value,BParam.DefGlyphFilter,'DefGlyph') then
     TryToSetFontDataParameter(S,Value,BParam.FD);
   end;
end;

{ TSXSkinEditStyle }

procedure TSXSkinEditStyle.AssignStyle(Style:TSXSkinStyle);
begin
 inherited;
 if Style is TSXSkinEditStyle then
  begin
   NUState:=TSXSkinEditStyle(Style).NUState;
   NFState:=TSXSkinEditStyle(Style).NFState;
   HUState:=TSXSkinEditStyle(Style).HUState;
   HFState:=TSXSkinEditStyle(Style).HFState;
   RUState:=TSXSkinEditStyle(Style).RUState;
   TextLeftOffset:=TSXSkinEditStyle(Style).TextLeftOffset;
   TextTopOffset:=TSXSkinEditStyle(Style).TextTopOffset;
   TextRightOffset:=TSXSkinEditStyle(Style).TextRightOffset;
   TextBottomOffset:=TSXSkinEditStyle(Style).TextBottomOffset;
   //
   HInEditEffect:=TSXSkinEditStyle(Style).HInEditEffect;
   HOutEditEffect:=TSXSkinEditStyle(Style).HOutEditEffect;
   EnableEditEffect:=TSXSkinEditStyle(Style).EnableEditEffect;
   DisableEditEffect:=TSXSkinEditStyle(Style).DisableEditEffect;
   FocusEditEffect:=TSXSkinEditStyle(Style).FocusEditEffect;
   UnfocusEditEffect:=TSXSkinEditStyle(Style).UnfocusEditEffect;
  end;
end;

procedure TSXSkinEditStyle.SetParameter(const Name,Value:String);
var EParam:PSXSkinEditStateParam;
      S,S2:String;
begin
 EParam:=nil;
 S:=Copy(Name,1,3);
 S2:=Copy(Name,1,4);
 if S='NU_' then EParam:=@NUState else
  if S='NF_' then EParam:=@NFState else
  if S='HU_' then EParam:=@HUState else
  if S='HF_' then EParam:=@HFState else
  if S='RU_' then EParam:=@RUState;
 if EParam=nil then
  begin
   if not TryToSetTransformEffectDataParameter(Name,Value,HInEditEffect,'HighlightInEdit') then
    if not TryToSetTransformEffectDataParameter(Name,Value,HOutEditEffect,'HighlightOutEdit') then
    if not TryToSetTransformEffectDataParameter(Name,Value,EnableEditEffect,'EnableEdit') then
    if not TryToSetTransformEffectDataParameter(Name,Value,DisableEditEffect,'DisableEdit') then
    if not TryToSetTransformEffectDataParameter(Name,Value,FocusEditEffect,'FocusEdit') then
    if not TryToSetTransformEffectDataParameter(Name,Value,UnfocusEditEffect,'UnfocusEdit') then
    if Name='TextLeftOffset' then
     TextLeftOffset:=StrToIntDef(Value,0) else
    if Name='TextTopOffset' then
     TextTopOffset:=StrToIntDef(Value,0) else
    if Name='TextRightOffset' then
     TextRightOffset:=StrToIntDef(Value,0) else
    if Name='TextBottomOffset' then
     TextBottomOffset:=StrToIntDef(Value,0);
  end else
   begin
    S:=Copy(Name,4,MaxInt);
    if S='Style' then EParam.Style:=Value else
     TryToSetFontDataParameter(S,Value,EParam.FD);
   end;
end;

{ TSXSkinCustomLibrary }

function TSXSkinCustomLibrary.GetCanBeUsed:Boolean;
begin
 Result:=FIsLoaded and FActive;
end;

procedure TSXSkinCustomLibrary.LoadFromFile(const FilePath:String);
var IniFile:TIniFile;
  SkinStyle:TSXSkinStyle;
     GStyle:TSXSkinGeneralStyle;
    Element:TSXSkinStyleElement;
 SL,SL2,SL3:TStringList;
      A,B,C:Integer;
       S,S2:String;
begin
 if FilePath='' then Styles.Clear;
 if not FileExists(FilePath) then exit;
 IniFile:=TIniFile.Create(FilePath);
 SL:=TStringList.Create;
 SL2:=TStringList.Create;
 try
  IniFile.ReadSections(SL);
  for A:=0 to SL.Count-1 do
   begin
    if SL[A]='BasedOn' then
     begin
      SL3:=TStringList.Create;
      try
       IniFile.ReadSectionValues(SL[A],SL3);
       for C:=0 to SL3.Count-1 do
        begin
         S:=WithLastSlash(FSkinDir)+ValueFromIndex(SL3,C);
         if FileExists(S) then
          LoadFromFile(S);
        end;
      finally
       SL3.Free;
      end;
     end else
    if SL[A]='Strings' then
     begin
      SL3:=TStringList.Create;
      try
       IniFile.ReadSectionValues(SL[A],SL3);
       for C:=0 to SL3.Count-1 do
        begin
         B:=Strings.IndexOfName(SL3.Names[C]);
         if B<0 then Strings.Add(SL3[C]) else
          SetValueFromIndex(Strings,B,ValueFromIndex(SL3,C));
        end;
      finally
       SL3.Free;
      end;
     end else
      begin
       SkinStyle:=nil;
       if Copy(SL[A],1,6)='_Label' then
        SkinStyle:=TSXSkinLabelStyle.Create else
       if Copy(SL[A],1,7)='_Button' then
        SkinStyle:=TSXSkinButtonStyle.Create else
       if Copy(SL[A],1,9)='_CheckBox' then
        SkinStyle:=TSXSkinCheckBoxStyle.Create else
       if Copy(SL[A],1,12)='_RadioButton' then
        SkinStyle:=TSXSkinRadioButtonStyle.Create else
       if Copy(SL[A],1,5)='_Edit' then
        SkinStyle:=TSXSkinEditStyle.Create else
       if Copy(SL[A],1,9)='_GroupBox' then
        SkinStyle:=TSXSkinGroupBoxStyle.Create else
       if Copy(SL[A],1,16)='_MultiStateCheck' then
        SkinStyle:=TSXSkinMultiStateCheckStyle.Create else
       if Copy(SL[A],1,11)='_MultiState' then
        SkinStyle:=TSXSkinMultiStateStyle.Create;
       if SkinStyle<>nil then
        begin
         SkinStyle.Name:=SL[A];
         Styles.AddUnique(SkinStyle);
         IniFile.ReadSectionValues(SL[A],SL2);
         for B:=0 to SL2.Count-1 do
          begin
           if SL2.Names[B]='BaseSkin' then
            begin
             C:=Styles.GetIndexByName(VarValue(ValueFromIndex(SL2,B)));
             if C>=0 then SkinStyle.AssignStyle(Styles[C]);
            end else SkinStyle.SetParameter(SL2.Names[B],VarValue(ValueFromIndex(SL2,B)));
          end;
        end else
         begin
          GStyle:=TSXSkinGeneralStyle.Create;
          GStyle.Name:=SL[A];
          Styles.AddUnique(GStyle);
          IniFile.ReadSectionValues(SL[A],SL2);
          for B:=0 to SL2.Count-1 do
           begin
            S:=VarValue(ValueFromIndex(SL2,B));
            if SL2.Names[B]='BaseSkin' then
             begin
              C:=Styles.GetIndexByName(S);
              if C>=0 then GStyle.AssignStyle(Styles[C]);
             end else
            if SL2.Names[B]='Transparent' then
             begin
              GStyle.Transparent:=S='1';
             end else
            if ((length(SL2.Names[B])=8) or (length(SL2.Names[B])=9)) and (Copy(SL2[B],1,7)='Element') then
             begin
              C:=Pos(',',S);
              if C>1 then
               begin
                S2:=Copy(S,C+1,MaxInt);
                Element:=CreateSkinStyleElementByType(S2);
                Element.Name:=Copy(S,1,C-1);
                GStyle.Elements.Add(Element);
               end;
             end else
              begin
               S2:=SL2.Names[B];
               C:=Pos('.',S2);
               if C>1 then S2:=Copy(S2,1,C-1);
               C:=GStyle.Elements.GetIndexByName(S2);
               if C>=0 then
                GStyle.Elements[C].SetProperty(Copy(SL2.Names[B],length(S2)+2,MaxInt),S,FilePath);
              end;
           end;
         end;
      end;
   end;
 finally
  IniFile.Free;
  SL.Free;
  SL2.Free;
 end;
end;

procedure TSXSkinCustomLibrary.SetActive(Value:Boolean);
begin
 if Value<>FActive then
  begin
   if not Value or (Value and ((FSkinFile='') or FileExists(GetFullSkinFile)) and
       ((FSkinFile2='') or FileExists(GetFullSkinFile2))) then
    begin
     FActive:=Value;
     if not (csLoading in ComponentState) then
      begin
       if Value then
        begin
         LoadFromFile(GetFullSkinFile);
         if FSkinFile2<>'' then
          LoadFromFile(GetFullSkinFile2);         
        end else LoadFromFile('');
       if Value or (csDesigning in ComponentState) then
        InvalidateControls;
      end;
    end;
  end;
end;

procedure TSXSkinCustomLibrary.SetSkinDir(const Value:String);
begin
 if Value<>FSkinDir then
  begin
   FSkinDir:=Value;
   if not (csLoading in ComponentState) then
    begin
     if FActive then
      begin
       LoadFromFile('');
       LoadFromFile(GetFullSkinFile);
       if FSkinFile2<>'' then
        LoadFromFile(GetFullSkinFile2);       
       InvalidateControls;
      end;
    end;
  end;
end;

procedure TSXSkinCustomLibrary.SetSkinFile(const Value:String);
begin
 if Value<>FSkinFile then
  begin
   FSkinFile:=Value;
   if not FileExists(GetFullSkinFile) then Active:=False;
   if not (csLoading in ComponentState) then
    begin
     if FActive then
      begin
       LoadFromFile('');
       LoadFromFile(GetFullSkinFile);
       if FSkinFile2<>'' then
        LoadFromFile(GetFullSkinFile2);               
       InvalidateControls;
      end;
    end;
  end;
end;

procedure TSXSkinCustomLibrary.SetSkinFile2(const Value:String);
begin
 if Value<>FSkinFile2 then
  begin
   FSkinFile2:=Value;
   if (Value<>'') and not FileExists(GetFullSkinFile2) then Active:=False;
   if not (csLoading in ComponentState) then
    begin
     if FActive then
      begin
       LoadFromFile('');
       LoadFromFile(GetFullSkinFile);
       if FSkinFile2<>'' then
        LoadFromFile(GetFullSkinFile2);               
       InvalidateControls;
      end;
    end;
  end;
end;

function TSXSkinCustomLibrary.VarValue(const Value:String):String;
begin
 Result:=Value;
 if (Value<>'') and (Value[1]='$') then
  Result:=GetStringsValue(Copy(Value,2,MaxInt));
end;

function TSXSkinCustomLibrary.GetFullSkinFile:String;
begin
 if FSkinFile='' then Result:='' else
 if PathIsRelative(FSkinFile) then
  begin
   if FSkinDir<>'' then
    Result:=GetFullPath(FSkinFile,WithLastSlash(FSkinDir)) else
     Result:=GetFullPath(FSkinFile,Application.ExeName);
  end else Result:=FSkinFile;
end;

function TSXSkinCustomLibrary.GetFullSkinFile2:String;
begin
 if FSkinFile2='' then Result:='' else
 if PathIsRelative(FSkinFile2) then
  begin
   if FSkinDir<>'' then
    Result:=GetFullPath(FSkinFile2,WithLastSlash(FSkinDir)) else
     Result:=GetFullPath(FSkinFile2,Application.ExeName);
  end else Result:=FSkinFile2;
end;

procedure TSXSkinCustomLibrary.AddControl(Control:TControl);
begin
 Controls.Add(Control);
end;

procedure TSXSkinCustomLibrary.RemoveControl(Control:TControl);
begin
 Controls.Remove(Control);
end;

procedure TSXSkinCustomLibrary.InvalidateControls;
var A:Integer;
begin
 for A:=0 to Controls.Count-1 do
  if TControl(Controls[A]) is TSXSkinCustomControl then
   TSXSkinCustomControl(Controls[A]).SkinChanged else
    TControl(Controls[A]).Invalidate;
end;

function TSXSkinCustomLibrary.GetStringsValue(const Name:String):String;
begin
 Result:=Strings.Values[Name];
end;

procedure TSXSkinCustomLibrary.Loaded;
begin
 inherited;
 FIsLoaded:=True;
 if FActive then
  begin
   LoadFromFile(GetFullSkinFile);
   if FSkinFile2<>'' then
    LoadFromFile(GetFullSkinFile2);       
   InvalidateControls;
  end;
end;

constructor TSXSkinCustomLibrary.Create(AOwner:TComponent);
begin
 inherited;
 Styles:=TSXSkinStyleList.Create;
 Controls:=TList.Create;
 Strings:=TStringList.Create;
end;

destructor TSXSkinCustomLibrary.Destroy;
begin
 Strings.Free;
 Styles.Free;
 Controls.Free;
 inherited Destroy;
end;

//////////////////////////////////////////////////////////////////////////

function CreateSkinStyleElementByType(const TypeName:String):TSXSkinStyleElement;
begin
 Result:=nil;
 if TypeName='IMAGE' then
  begin
   Result:=TSXSkinStyleImageElement.Create;
   TSXSkinStyleImageElement(Result).ImageType:=ssitAutoDetect;
  end else
 if TypeName='PNG_IMAGE' then
  begin
   Result:=TSXSkinStyleImageElement.Create;
   TSXSkinStyleImageElement(Result).ImageType:=ssitPNG;
  end else
 if TypeName='JPEG_IMAGE' then
  begin
   Result:=TSXSkinStyleImageElement.Create;
   TSXSkinStyleImageElement(Result).ImageType:=ssitJPEG;
  end else
 if TypeName='BOX_TILE' then
  begin
   Result:=TSXSkinStyleBoxTileElement.Create;
  end else
 if TypeName='RECT' then
  begin
   Result:=TSXSkinStyleFigureElement.Create;
   TSXSkinStyleFigureElement(Result).FigureType:=ssftRectangle;
  end else
 if TypeName='CIRCLE' then
  begin
   Result:=TSXSkinStyleFigureElement.Create;
   TSXSkinStyleFigureElement(Result).FigureType:=ssftCircle;
  end else
 if TypeName='ROUND_RECT' then
  begin                                          
   Result:=TSXSkinStyleFigureElement.Create;
   TSXSkinStyleFigureElement(Result).FigureType:=ssftRoundRectangle;
  end else
 if TypeName='LINE' then
  begin
   Result:=TSXSkinStyleFigureElement.Create;
   TSXSkinStyleFigureElement(Result).FigureType:=ssftLine;
   TSXSkinStyleFigureElement(Result).HasBorder:=True;
  end else
 if TypeName='SOLID_FILL' then
  begin
   Result:=TSXSkinStyleFigureElement.Create;
   TSXSkinStyleFigureElement(Result).FigureType:=ssftSolidFill;
  end else
 if TypeName='FOCUS_RECT' then
  begin
   Result:=TSXSkinStyleFigureElement.Create;
   TSXSkinStyleFigureElement(Result).FigureType:=ssftFocusRectangle;
  end;
 if Result=nil then Result:=TSXSkinStyleElement.Create;
end;

function GetFilterTypeByName(const Name:String):TSXImageFilterType;
begin
 Result:=iftNone;
 if Name='Lighten' then Result:=iftLighten else
  if Name='LightenHorizG' then Result:=iftLightenHorizG else
  if Name='LightenVertG' then Result:=iftLightenVertG else
  if Name='Darken' then Result:=iftDarken else
  if Name='DarkenHorizG' then Result:=iftDarkenHorizG else
  if Name='DarkenVertG' then Result:=iftDarkenVertG else
  if Name='Alpha' then Result:=iftAlpha else
  if Name='Monochrome' then Result:=iftMonochrome else
  if Name='ColorOverlay' then Result:=iftColorOverlay else
  if Name='ColorOverlayHorizG' then Result:=iftColorOverlayHorizG else
  if Name='ColorOverlayVertG' then Result:=iftColorOverlayVertG;
end;

function HasTransformEffect(const Effect:TSXTransformEffectData):Boolean;
begin
 Result:=((Effect.OldType<>tetNone) or (Effect.NewType<>tetNone)) and (Effect.StepsNum>1);
end;

function GetTransformEffectTypeByName(const Name:String):TSXTransformEffectType;
begin
 Result:=tetNone;
 if Name='Blend' then Result:=tetBlend else
  if Name='Clear' then Result:=tetClear else
  if Name='Fade' then Result:=tetFade else
  if Name='OverDraw' then Result:=tetOverDraw else
  if Name='Slide' then Result:=tetSlide;
end;

function GetTransformEffectDirectionByName(const Name:String):TSXTransformEffectDirection;
begin
 Result:=tedLeft;
 if Name='Right' then Result:=tedRight else
  if Name='Top' then Result:=tedTop else
  if Name='Bottom' then Result:=tedBottom else
  if Name='LeftRight' then Result:=tedLeftRight else
  if Name='TopBottom' then Result:=tedTopBottom else
  if Name='LeftTop' then Result:=tedLeftTop else
  if Name='LeftBottom' then Result:=tedLeftBottom else
  if Name='RightTop' then Result:=tedRightTop else
  if Name='RightBottom' then Result:=tedRightBottom;
end;

procedure GetTransformEffectByName(const Name:String;var Effect:TSXTransformEffectData);
var EType:(etNone,
           //Slide&Slide
           etSlideLeft,etSlideTop,etSlideRight,etSlideBottom,etSlideLeftTop,
           etSlideLeftBottom,etSlideRightTop,etSlideRightBottom,
           //OverDraw&OverDraw
           etOverDrawLeft,etOverDrawTop,etOverDrawRight,etOverDrawBottom,
           etOverDrawLeftTop,etOverDrawLeftBottom,etOverDrawRightTop,
           etOverDrawRightBottom,etOverDrawLeftRight,etOverDrawTopBottom,
           etOverDrawLeftRightOut,etOverDrawTopBottomOut);
begin
 EType:=etNone;
 if Name='None' then
  begin
   Effect.OldType:=tetNone;
   Effect.NewType:=tetNone;
   exit;
  end;
 if Name='SlideLeft' then EType:=etSlideLeft else
  if Name='SlideTop' then EType:=etSlideTop else
  if Name='SlideRight' then EType:=etSlideRight else
  if Name='SlideBottom' then EType:=etSlideBottom else
  if Name='SlideLeftTop' then EType:=etSlideLeftTop else
  if Name='SlideLeftBottom' then EType:=etSlideLeftBottom else
  if Name='SlideRightTop' then EType:=etSlideRightTop else
  if Name='SlideRightBottom' then EType:=etSlideRightBottom else
  if Name='OverDrawLeft' then EType:=etOverDrawLeft else
  if Name='OverDrawTop' then EType:=etOverDrawTop else
  if Name='OverDrawRight' then EType:=etOverDrawRight else
  if Name='OverDrawBottom' then EType:=etOverDrawBottom else
  if Name='OverDrawLeftTop' then EType:=etOverDrawLeftTop else
  if Name='OverDrawLeftBottom' then EType:=etOverDrawLeftBottom else
  if Name='OverDrawRightTop' then EType:=etOverDrawRightTop else
  if Name='OverDrawRightBottom' then EType:=etOverDrawRightBottom else
  if Name='OverDrawLeftRight' then EType:=etOverDrawLeftRight else
  if Name='OverDrawTopBottom' then EType:=etOverDrawTopBottom else
  if Name='OverDrawLeftRightOut' then EType:=etOverDrawLeftRightOut else
  if Name='OverDrawTopBottomOut' then EType:=etOverDrawTopBottomOut;
 if EType in [etSlideLeft,etSlideTop,etSlideRight,etSlideBottom,etSlideLeftTop,
              etSlideLeftBottom,etSlideRightTop,etSlideRightBottom] then
  begin
   Effect.OldType:=tetSlide;
   Effect.OldDirOut:=True;
   Effect.NewType:=tetSlide;
   case EType of
    etSlideLeft:        begin
                         Effect.OldDirection:=tedLeft;
                         Effect.NewDirection:=tedRight;
                         Effect.Offset:=10;
                        end;
    etSlideTop:         begin
                         Effect.OldDirection:=tedTop;
                         Effect.NewDirection:=tedBottom;
                         Effect.Offset:=5;
                        end;
    etSlideRight:       begin
                         Effect.OldDirection:=tedRight;
                         Effect.NewDirection:=tedLeft;
                         Effect.Offset:=10;
                        end;
    etSlideBottom:      begin
                         Effect.OldDirection:=tedBottom;
                         Effect.NewDirection:=tedTop;
                         Effect.Offset:=5;
                        end;
    etSlideLeftTop:     begin
                         Effect.OldDirection:=tedLeftTop;
                         Effect.NewDirection:=tedRightBottom;
                         Effect.Offset:=1;
                        end;
    etSlideLeftBottom:  begin
                         Effect.OldDirection:=tedLeftBottom;
                         Effect.NewDirection:=tedRightTop;
                         Effect.Offset:=1;
                        end;
    etSlideRightTop:    begin
                         Effect.OldDirection:=tedRightTop;
                         Effect.NewDirection:=tedLeftBottom;
                         Effect.Offset:=1;
                        end;
    etSlideRightBottom: begin
                         Effect.OldDirection:=tedRightBottom;
                         Effect.NewDirection:=tedLeftTop;
                         Effect.Offset:=1;
                        end;
   end;
  end else
 if EType in [etOverDrawLeft,etOverDrawTop,etOverDrawRight,etOverDrawBottom,
              etOverDrawLeftTop,etOverDrawLeftBottom,etOverDrawRightTop,
              etOverDrawRightBottom,etOverDrawLeftRight,etOverDrawTopBottom,
              etOverDrawLeftRightOut,etOverDrawTopBottomOut] then
  begin
   Effect.OldType:=tetOverDraw;
   Effect.OldDirOut:=True;
   Effect.NewType:=tetOverDraw;
   case EType of
    etOverDrawLeft:         begin
                             Effect.OldDirection:=tedLeft;
                             Effect.NewDirection:=tedRight;
                            end;
    etOverDrawTop:          begin
                             Effect.OldDirection:=tedTop;
                             Effect.NewDirection:=tedBottom;
                            end;
    etOverDrawRight:        begin
                             Effect.OldDirection:=tedRight;
                             Effect.NewDirection:=tedLeft;
                            end;
    etOverDrawBottom:       begin
                             Effect.OldDirection:=tedBottom;
                             Effect.NewDirection:=tedTop;
                            end;
    etOverDrawLeftTop:      begin
                             Effect.OldDirection:=tedLeftTop;
                             Effect.NewDirection:=tedLeftTop;
                             Effect.NewDirOut:=True;
                             Effect.NewInverted:=True;
                            end;
    etOverDrawLeftBottom:   begin
                             Effect.OldDirection:=tedLeftBottom;
                             Effect.NewDirection:=tedLeftBottom;
                             Effect.NewDirOut:=True;
                             Effect.NewInverted:=True;
                            end;
    etOverDrawRightTop:     begin
                             Effect.OldDirection:=tedRightTop;
                             Effect.NewDirection:=tedRightTop;
                             Effect.NewDirOut:=True;
                             Effect.NewInverted:=True;
                            end;
    etOverDrawRightBottom:  begin
                             Effect.OldDirection:=tedRightBottom;
                             Effect.NewDirection:=tedRightBottom;
                             Effect.NewDirOut:=True;
                             Effect.NewInverted:=True;
                            end;
    etOverDrawLeftRight:    begin
                             Effect.OldDirection:=tedLeftRight;
                             Effect.OldDirOut:=False;
                             Effect.OldInverted:=True;
                             Effect.NewDirection:=tedLeftRight;
                            end;
    etOverDrawTopBottom:    begin
                             Effect.OldDirection:=tedTopBottom;
                             Effect.OldDirOut:=False;
                             Effect.OldInverted:=True;
                             Effect.NewDirection:=tedTopBottom;
                            end;
    etOverDrawLeftRightOut: begin
                             Effect.OldDirection:=tedLeftRight;
                             Effect.NewDirection:=tedLeftRight;
                             Effect.NewDirOut:=True;
                             Effect.NewInverted:=True;
                            end;
    etOverDrawTopBottomOut: begin
                             Effect.OldDirection:=tedTopBottom;
                             Effect.NewDirection:=tedTopBottom;
                             Effect.NewDirOut:=True;
                             Effect.NewInverted:=True;
                            end;
   end;
  end;
end;

procedure ApplyFilterToBitmap(var Bitmap:TBitmap32;const Filter:TSXFilterData);
begin
 case Filter.AType of
  iftLighten:            Lighten(Bitmap,Filter.Color);
  iftLightenHorizG:      LightenHorizG(Bitmap,Filter.Color,Filter.Color2);
  iftLightenVertG:       LightenVertG(Bitmap,Filter.Color,Filter.Color2);
  iftDarken:             Darken(Bitmap,Filter.Color);
  iftDarkenHorizG:       DarkenHorizG(Bitmap,Filter.Color,Filter.Color2);
  iftDarkenVertG:        DarkenVertG(Bitmap,Filter.Color,Filter.Color2);
  iftAlpha:              MultiplyAlpha(Bitmap,Filter.Value);
  iftMonochrome:         Monochrome(Bitmap,Filter.Color,Filter.Color2);
  iftColorOverlay:       ColorOverlay(Bitmap,Filter.Color);
  iftColorOverlayHorizG: ColorOverlayHorizG(Bitmap,Filter.Color,Filter.Color2);
  iftColorOverlayVertG:  ColorOverlayVertG(Bitmap,Filter.Color,Filter.Color2);
 end;
end;

procedure ApplyTransformEffectToBitmaps(B1,B2:TBitmap32;const Effect:TSXTransformEffectData;
           CurrentStep:Integer;var B:TBitmap32);
var A:Integer;

 procedure DrawBitmap(New:Boolean;CurrentStep:Integer);
 var   AType:TSXTransformEffectType;
  ADirection:TSXTransformEffectDirection;
     ADirOut:Boolean;
     AInvert:Boolean;
      Bitmap:TBitmap32;
         X,Y:Integer;
        R,R1:TRect;
 begin
  if New then
   begin
    AType:=Effect.NewType;
    ADirection:=Effect.NewDirection;
    ADirOut:=Effect.NewDirOut;
    AInvert:=Effect.NewInverted;
    Bitmap:=B2;
   end else
    begin
     AType:=Effect.OldType;
     ADirection:=Effect.OldDirection;
     ADirOut:=Effect.OldDirOut;
     AInvert:=Effect.OldInverted;
     Bitmap:=B1;
    end;
  case AType of
   tetNone:     Bitmap.DrawTo(B);
   tetClear:    ;
   tetFade:     begin
                 if New then
                  A:=RcTable[Effect.StepsNum,CurrentStep] else
                   A:=255-RcTable[Effect.StepsNum,CurrentStep];
                 DrawToAsAlpha(Bitmap,B,0,0,A);
                end;
   tetOverDraw: begin
                 if ADirOut then CurrentStep:=Effect.StepsNum-CurrentStep;
                 if ADirection=tedLeftRight then
                  begin
                   X:=round((B.Width+Effect.Offset+0.1)/Effect.StepsNum*CurrentStep/2);
                   if AInvert then
                    begin
                     R1:=Rect(X,0,B.Width-X,B.Height);
                     Bitmap.DrawTo(B,R1.Left,R1.Top,R1);
                    end else
                     begin
                      R1:=Rect(0,0,X,B.Height);
                      Bitmap.DrawTo(B,R1.Left,R1.Top,R1);
                      R1:=Rect(B.Width-X,0,B.Width,B.Height);
                      Bitmap.DrawTo(B,R1.Left,R1.Top,R1);
                     end;
                  end else
                 if ADirection=tedTopBottom then
                  begin
                   Y:=round((B.Height+Effect.Offset+0.1)/Effect.StepsNum*CurrentStep/2);
                   if AInvert then
                    begin
                     R1:=Rect(0,Y,B.Width,B.Height-Y);
                     Bitmap.DrawTo(B,R1.Left,R1.Top,R1);
                    end else
                     begin
                      R1:=Rect(0,0,B.Width,Y);
                      Bitmap.DrawTo(B,R1.Left,R1.Top,R1);
                      R1:=Rect(0,B.Height-Y,B.Width,B.Height);
                      Bitmap.DrawTo(B,R1.Left,R1.Top,R1);
                     end;
                  end else
                   begin
                    R:=Rect(0,0,B.Width,B.Height);
                    case ADirection of
                     tedLeft,
                     tedLeftTop,
                     tedLeftBottom:  R.Right:=round((B.Width+Effect.Offset+0.1)/Effect.StepsNum*CurrentStep)-Effect.Offset;
                     tedRight,
                     tedRightTop,
                     tedRightBottom: R.Left:=-round((B.Width+Effect.Offset+0.1)/Effect.StepsNum*CurrentStep)+B.Width+Effect.Offset;
                    end;
                    case ADirection of
                     tedTop,
                     tedLeftTop,
                     tedRightTop:    R.Bottom:=round((B.Height+Effect.Offset+0.1)/Effect.StepsNum*CurrentStep)-Effect.Offset;
                     tedBottom,
                     tedLeftBottom,
                     tedRightBottom: R.Top:=-round((B.Height+Effect.Offset+0.1)/Effect.StepsNum*CurrentStep)+B.Height+Effect.Offset;
                    end;
                    if AInvert then
                     begin
                      R1:=Rect(0,0,B.Width,R.Top);
                      if (R1.Right>R1.Left) and (R1.Bottom>R1.Top) then
                       Bitmap.DrawTo(B,R1.Left,R1.Top,R1);
                      //
                      R1:=Rect(0,R.Top,R.Left,R.Bottom);
                      if (R1.Right>R1.Left) and (R1.Bottom>R1.Top) then
                       Bitmap.DrawTo(B,R1.Left,R1.Top,R1);
                      //
                      R1:=Rect(R.Right,R.Top,B.Width,R.Bottom);
                      if (R1.Right>R1.Left) and (R1.Bottom>R1.Top) then
                       Bitmap.DrawTo(B,R1.Left,R1.Top,R1);
                      //
                      R1:=Rect(0,R.Bottom,B.Width,B.Height);
                      if (R1.Right>R1.Left) and (R1.Bottom>R1.Top) then
                       Bitmap.DrawTo(B,R1.Left,R1.Top,R1);
                     end else Bitmap.DrawTo(B,R.Left,R.Top,R);
                   end;
                end;
   tetSlide:    begin
                 if not ADirOut then CurrentStep:=Effect.StepsNum-CurrentStep;
                 if ADirection=tedLeftRight then
                  begin
                   X:=round((B.Width+Effect.Offset+0.1)/Effect.StepsNum*CurrentStep/2);
                   Bitmap.DrawTo(B,-X,0,Rect(0,0,B.Width div 2,B.Height));
                   Bitmap.DrawTo(B,(B.Width div 2)+X,0,Rect(B.Width div 2,0,B.Width,B.Height));
                  end else
                 if ADirection=tedTopBottom then
                  begin
                   Y:=round((B.Height+Effect.Offset+0.1)/Effect.StepsNum*CurrentStep/2);
                   Bitmap.DrawTo(B,0,-Y,Rect(0,0,B.Width,B.Height div 2));
                   Bitmap.DrawTo(B,0,(B.Height div 2)+Y,Rect(0,B.Height div 2,B.Width,B.Height));
                  end else
                   begin
                    X:=0; Y:=0;
                    case ADirection of
                     tedLeft,
                     tedLeftTop,
                     tedLeftBottom:  X:=-round((B.Width+Effect.Offset+0.1)/Effect.StepsNum*CurrentStep);
                     tedRight,
                     tedRightTop,
                     tedRightBottom: X:=round((B.Width+Effect.Offset+0.1)/Effect.StepsNum*CurrentStep);
                    end;
                    case ADirection of
                     tedTop,
                     tedLeftTop,
                     tedRightTop:    Y:=-round((B.Height+Effect.Offset+0.1)/Effect.StepsNum*CurrentStep);
                     tedBottom,
                     tedLeftBottom,
                     tedRightBottom: Y:=round((B.Height+Effect.Offset+0.1)/Effect.StepsNum*CurrentStep);
                    end;
                    Bitmap.DrawTo(B,X,Y);
                   end;
                end;
  end;
 end;

begin
 if Effect.NewType=tetBlend then
  begin
   A:=RcTable[Effect.StepsNum,CurrentStep];
   DrawMixedBitmap(B,B1,B2,A);
   exit;
  end;
 DrawBitmap(Effect.OldOnTop,CurrentStep);
 DrawBitmap(not Effect.OldOnTop,CurrentStep);
end;

end.
